from __future__ import annotations

import logging

from slack_sdk.models.blocks.basic_components import Option, TextObject
from slack_sdk.models.blocks.block_elements import StaticSelectElement
from slack_sdk.models.blocks.blocks import (
    ActionsBlock,
    Block,
    ContextBlock,
    SectionBlock,
)
from slack_sdk.models.views import View

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.models.incident import Incident
from firefighter.slack.slack_templating import slack_block_footer, slack_block_separator
from firefighter.slack.views.modals.base_modal.base import SlackModal
from firefighter.slack.views.modals.base_modal.mixins import (
    IncidentSelectableModalMixinBase,
)

logger = logging.getLogger(__name__)


class SelectModal(SlackModal):
    handle_modal_fn = None

    callback_id = "select_incident"

    def build_modal_fn(
        self,
        incident: Incident | None = None,
        callback_id: str | None = "update",
        select_class: IncidentSelectableModalMixinBase | None = None,
        *,
        missing_incident_error: bool = False,
    ) -> View:
        active_incidents = Incident.objects.filter(
            _status__lt=IncidentStatus.CLOSED
        ).order_by("-id")

        if select_class is None and isinstance(self, IncidentSelectableModalMixinBase):
            select_class = self
        elif select_class is None or not isinstance(
            select_class, IncidentSelectableModalMixinBase
        ):
            select_class = (
                self
                if isinstance(self, IncidentSelectableModalMixinBase)
                else IncidentSelectableModalMixinBase()
            )

        incident_options = [
            Option(
                value=str(inc.id),
                label=f"{inc.id}: {inc.title}"[:73],
            )
            for inc in active_incidents
        ]
        if len(incident_options) > 0:
            select_incident_blocks: list[Block] = [
                SectionBlock(text=select_class.get_select_title()),
                ActionsBlock(
                    block_id="incident_update_select_incident",
                    elements=[
                        StaticSelectElement(
                            action_id="incident_update_select_incident",
                            placeholder="Select an active incident",
                            options=incident_options,
                        ),
                    ],
                ),
            ]
        else:
            select_incident_blocks = [
                SectionBlock(
                    text=":x: This command requires an active incident. Please create one first."
                )
            ]
        if missing_incident_error:
            select_incident_blocks.append(
                ContextBlock(
                    elements=[TextObject(text=":x: Please select an incident!")]
                )
            )
        select_incident_blocks.extend((slack_block_separator(), slack_block_footer()))

        return View(
            type="modal",
            title=select_class.get_select_modal_title()[:24],
            callback_id=callback_id,
            private_metadata=str(incident.id) if incident else "",
            blocks=select_incident_blocks,
        )


modal_select = SelectModal()



from __future__ import annotations

from django.apps import AppConfig


class JiraAppConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    label = "jira_app"
    name = "firefighter.jira_app"
    verbose_name = "Jira"

    def ready(self) -> None:
        import firefighter.jira_app.tasks

        return super().ready()



from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING, cast

from celery import shared_task
from django.db.utils import IntegrityError

from firefighter.confluence.models import Runbook
from firefighter.confluence.service import confluence_service

if TYPE_CHECKING:
    from firefighter.confluence.utils import ConfluencePageId

logger = logging.getLogger(__name__)

ALLOWED_TYPES = {
    "ms",
    "front",
    "infra",
    "app",
    "other",
    "task",
    "stream",
    "site",
    "msf",
    "data",
    "gw",
    "nbk",
    "external",
}


@shared_task(name="confluence.sync_runbooks")
def sync_runbooks() -> None:
    all_fetched_ids: set[ConfluencePageId] = set()
    folders = confluence_service.get_page_children_pages(
        confluence_service.RUNBOOKS_FOLDER_ID
    )

    for folder in folders:
        runbooks_pages = confluence_service.get_page_children_pages(folder["id"])
        for page in runbooks_pages:
            data = confluence_service.parse_confluence_page(page)
            page_id = data["page_id"]
            all_fetched_ids.add(page_id)
            data["name"] = data["name"].removesuffix("[RUNBOOK]").strip()

            data_editable = cast("dict[str, str]", data)
            data_editable["title"] = (
                data_editable["name"].removesuffix("[RUNBOOK]").strip()
            )
            # Remove everything between braces
            cleaned_title = re.sub(r"\[[^\]]+\]", "", data_editable["title"])
            data_editable["service_type"] = cleaned_title.split(".", maxsplit=1)[
                0
            ].strip()
            data_editable["service_name"] = (
                cleaned_title.split(".", maxsplit=1)[1]
                if "." in cleaned_title
                else cleaned_title
            )

            if data_editable["service_type"] not in ALLOWED_TYPES:
                data_editable["service_type"] = "other"
            try:
                Runbook.objects.update_or_create(
                    page_id=int(data_editable.pop("page_id")), defaults=data_editable
                )
            except IntegrityError:
                logger.exception(f"IntegrityError for {data} {data}")
    # Print all runbooks that are not in the folder anymore
    missing_runbooks = Runbook.objects.exclude(page_id__in=all_fetched_ids)
    if not missing_runbooks:
        logger.info(f"Synced {len(all_fetched_ids)} runbooks. No runbooks to delete.")
        return
    if len(missing_runbooks) > 4:
        logger.warning(
            f"Too many runbooks not found. FireFighter will not delete them automatically, as it might be an API or implementation error. List: {[f'{x.page_id}/ {x.name}' for x in missing_runbooks]}"
        )
        return
    logger.warning(
        f"Missing runbooks that will be deleted: {[f'{x.page_id}/ {x.name}' for x in missing_runbooks]}"
    )
    # Delete all runbooks that are not in the folders anymore
    missing_runbooks.delete()
    return



from __future__ import annotations

import logging
import operator

from celery import shared_task
from django.utils.timezone import (  # type: ignore[attr-defined]
    datetime,
    get_current_timezone,
    timedelta,
)

from firefighter.confluence.service import confluence_service
from firefighter.confluence.utils import (
    CONFLUENCE_PM_ARCHIVE_TITLE_REGEX,
    CONFLUENCE_PM_TITLE_REGEX,
    ConfluencePage,
    ConfluencePageId,
    parse_postmortem_title,
)
from firefighter.firefighter.utils import get_in
from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.models.incident import Incident
from firefighter.incidents.views.date_utils import get_quarter_from_week

logger = logging.getLogger(__name__)


@shared_task(
    name="confluence.archive_and_sort_postmortems",
    soft_time_limit=60 * 5,
    time_limit=60 * 7,
)
def archive_and_sort_postmortems(*, dry_run: bool = True) -> None:
    """Archive postmortems that are older than 2 weeks, for incidents that are closed.

    Args:
        dry_run (bool, optional): Should actions be performed? If True, Confluence will be accessed in read-only mode. Defaults to False.
    """
    logger.info(f"Running in {'dry run' if dry_run else 'real'} mode")

    # 1. Get top level pages
    pm_to_sort, quarter_bins = _get_top_level_pages(
        confluence_service.POSTMORTEM_FOLDER_ID
    )

    # 2. Create current quarter folder if it does not exist
    create_current_bin_if_needed(quarter_bins, dry_run=dry_run)

    # 3. Archive postmortems in root into their quarter folder (if appropriate)
    _archive_postmortems(pm_to_sort, quarter_bins, dry_run=dry_run)

    # 4. Move archived postmortems in the right quarter folder if there is any error
    # Not needed atm

    # 5. Sort postmortems in each quarter folder
    sort_postmortems_in_bins(quarter_bins, dry_run=dry_run)


def _get_top_level_pages(
    root_page_id: ConfluencePageId,
) -> tuple[list[ConfluencePage], dict[str, int]]:
    """Get all top level pages in the postmortem folder."""
    pm_to_sort: list[ConfluencePage] = []
    quarter_bins: dict[str, int] = {}
    children_pages = confluence_service.get_page_children_pages(root_page_id)

    for page in children_pages:
        title: str = page["title"]
        page_id: int = int(page["id"])

        pm_reg = CONFLUENCE_PM_TITLE_REGEX.match(title)
        if pm_reg:
            pm_to_sort.append(page)
            continue

        quarter_bin = CONFLUENCE_PM_ARCHIVE_TITLE_REGEX.search(title)
        if not quarter_bin:
            if "[Template]" not in title:
                logger.warning(f"Could not match `{title}`, ID: {page_id}")
            continue
        year = quarter_bin.group(2)
        quarter = quarter_bin.group(1)
        fmt = f"{year}Q{quarter}"
        if fmt in quarter_bins:
            logger.warning(f"Found duplicate quarter bin: {fmt}")
        quarter_bins[fmt] = page_id
        logger.debug(f"Found quarter {fmt} with id {page_id}")

    logger.info(f"Found {len(pm_to_sort)} postmortems to sort in root folder.")
    logger.info(f"Found {len(quarter_bins)} quarters bins ({quarter_bins})")
    return pm_to_sort, quarter_bins


def _archive_postmortems(
    pm_to_sort: list[ConfluencePage],
    quarter_bins: dict[str, int],
    *,
    dry_run: bool = False,
) -> None:
    """Move postmortems to their quarter bin, if they have a related incident in FireFighter that is closed for more than 7 days.

    Args:
        pm_to_sort (list[ConfluencePage]): Confluence pages to sort. Not Django models, but the JSON response from the API.
        quarter_bins (dict[str, int]): Mapping of quarter/year to Confluence page ID.
        dry_run (bool, optional): Defaults to False.
    """
    for pm in pm_to_sort:
        if not pm["title"].startswith("#"):
            logger.debug(f"Skipping {pm['title']} because it does not look like a PM")
            continue

        date_, fmt = parse_postmortem_title(pm["title"])
        if not date_ or not fmt:
            logger.error(f"Could not parse {pm['title']} from page id {pm['id']}")
            continue
        if fmt not in quarter_bins:
            logger.error(f"Could not find quarter page {fmt} for {pm['title']}")
            continue

        incident = Incident.objects.filter(postmortem_for__page_id=pm["id"]).first()
        if not incident:
            logger.warning(
                f"Could not find incident for `{pm['title']}` ({pm['id']}). Is it linked in FireFighter and has it the correct title format?"
            )
            continue
        if incident.status != IncidentStatus.CLOSED:
            logger.debug(f"Skipping `{pm['title']}` because incident is not closed")
            continue
        latest_closed_status = (
            incident.incidentupdate_set.filter(_status=IncidentStatus.CLOSED)
            .order_by("-created_at")
            .first()
        )
        if not latest_closed_status:
            logger.warning(
                f"Could not find latest closed status for incident #{incident.id}."
            )
            continue
        if latest_closed_status.created_at >= datetime.now(
            tz=get_current_timezone()
        ) - timedelta(days=7):
            logger.debug(
                f"Skipping `{pm['title']}` because incident is not closed for more than 7 days"
            )
            continue

        confluence_service.move_page(pm["id"], quarter_bins[fmt], dry_run=dry_run)


def create_current_bin_if_needed(
    quarter_bins: dict[str, int], *, dry_run: bool = False
) -> bool:
    current_quarter = get_quarter_from_week(datetime.now().isocalendar()[1])
    current_year = datetime.now().isocalendar()[0]
    current_quarter_year = f"{current_year}Q{current_quarter}"
    if current_quarter_year not in quarter_bins:
        page_id = create_current_bin(
            quarter_bins, current_quarter, current_year, dry_run=dry_run
        )
        quarter_bins[current_quarter_year] = page_id
        return True
    return False


def create_current_bin(
    quarter_bins: dict[str, int],
    current_quarter: int,
    current_year: int,
    *,
    dry_run: bool = False,
) -> int:
    res = confluence_service.create_page(
        f"Q{current_quarter} {current_year} Postmortems",
        confluence_service.POSTMORTEM_FOLDER_ID,
        body="",
    )
    page_id = int(get_in(res, "id"))
    confluence_service.move_page(
        page_id,
        next(iter(quarter_bins.keys())),
        position="before",
        dry_run=dry_run,
    )

    return page_id


def sort_postmortems_in_bins(
    quarter_bins: dict[str, int], *, dry_run: bool = False
) -> None:
    """Sort postmortems in each quarter archive folder.

    Args:
        quarter_bins (dict[str, int]): _description_
        dry_run (bool, optional): _description_. Defaults to False.
    """
    for quarter, quarter_page_id in quarter_bins.items():
        # Get children per archive page
        children: list[ConfluencePage] = confluence_service.get_page_children_pages(
            quarter_page_id, expand=""
        )
        logger.info(f"Found {len(children)} postmortems in {quarter}")
        quarter_children: list[tuple[int, datetime, ConfluencePage]] = []

        for page in children:
            title: str = page["title"]
            page_id: int = int(page["id"])

            date_, fmt = parse_postmortem_title(title)
            if date_ and fmt:
                quarter_children.append((page_id, date_, page))

        sorted_children = sorted(quarter_children, key=operator.itemgetter(1), reverse=True)
        # Check if already sorted
        if sorted_children == quarter_children:
            logger.debug(f"Quarter {quarter} is already sorted. Skipping")
            continue

        logger.info(f"Quarter {quarter} is not sorted. Sorting")

        clean_sorted: list[tuple[ConfluencePageId, ConfluencePage]] = [
            (x[0], x[2]) for x in sorted_children
        ]

        confluence_service.sort_pages(clean_sorted, dry_run=dry_run)

        logger.info(f"Fetched {len(quarter_children)} postmortems")



"""Hatchling build hook to run JS and CSS build steps with Rollup."""

from __future__ import annotations

import subprocess  # noqa: S404
from typing import Any

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class RollupBuildHook(BuildHookInterface):
    def initialize(self, version: str, build_data: dict[str, Any]) -> Any:
        subprocess.check_output("npm install", shell=True)  # noqa: S602,S607
        subprocess.check_output("npm run build", shell=True)  # noqa: S602,S607
        return super().initialize(version, build_data)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from slack_sdk.models.blocks.block_elements import ButtonElement
from slack_sdk.models.blocks.blocks import ActionsBlock
from slack_sdk.models.views import View

from firefighter.slack.views.modals.base_modal.base import SlackModal
from firefighter.slack.views.modals.base_modal.mixins import (
    IncidentSelectableModalMixin,
)
from firefighter.slack.views.modals.edit import EditMetaModal
from firefighter.slack.views.modals.update_roles import UpdateRolesModal
from firefighter.slack.views.modals.update_status import UpdateStatusModal

if TYPE_CHECKING:
    from firefighter.incidents.models.incident import Incident

logger = logging.getLogger(__name__)


class UpdateModal(
    IncidentSelectableModalMixin,
    SlackModal,
):
    handle_modal_fn = None
    open_action: str = "update_incident"
    callback_id: str = "incident_update"

    def build_modal_fn(self, incident: Incident, **kwargs: Any) -> View:
        return View(
            type="modal",
            title=f"Update incident #{incident.id}"[:24],
            callback_id=self.callback_id,
            private_metadata=str(incident.id),
            blocks=[
                ActionsBlock(
                    block_id="choose_update_type",
                    elements=[
                        ButtonElement(
                            text="Update status",
                            action_id=UpdateStatusModal.push_action,
                        ),
                        ButtonElement(
                            text="Update roles",
                            action_id=UpdateRolesModal.push_action,
                        ),
                        ButtonElement(
                            text="Edit meta",
                            action_id=EditMetaModal.push_action,
                        ),
                    ],
                )
            ],
        )

    def get_select_modal_title(self) -> str:
        return "Update incident"

    def get_select_title(self) -> str:
        return "Select a critical incident to update"


modal_update = UpdateModal()



"""This is the main entrypoint for FireFighter.

!!! warning

    This app does not contain any business logic. It is responsible for loading all the settings, providing common base features and tying together all the other apps.

## Features

- Load all settings
- Provide `ASGI` and `WSGI` entrypoints
- Configure loggers
- Provide a Celery app
- Set the base URL routing
- Provides a `healthcheck` endpoint
- Provide a `robots.txt` endpoint
- Provides some utils
- Provides SSO integration
- Register the Django admin and customize its theme

"""

from __future__ import annotations

from firefighter.firefighter.celery_client import app as celery_app

__all__ = ("celery_app",)



from __future__ import annotations

import logging
import uuid
from typing import TYPE_CHECKING, Any

from django.db import models
from django.db.models.signals import pre_save
from django.dispatch import receiver
from django.utils import timezone
from django_stubs_ext.db.models import TypedModelMeta

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.models.component import Component
from firefighter.incidents.models.environment import Environment
from firefighter.incidents.models.priority import Priority
from firefighter.incidents.models.severity import Severity
from firefighter.incidents.models.user import User

if TYPE_CHECKING:
    import datetime

    from firefighter.incidents.models.incident import Incident  # noqa: F401

logger = logging.getLogger(__name__)


class IncidentUpdateManager(models.Manager["IncidentUpdate"]):
    def get_or_none(self, **kwargs: Any) -> IncidentUpdate | None:
        try:
            return self.get(**kwargs)
        except IncidentUpdate.DoesNotExist:
            return None


class IncidentUpdate(models.Model):
    """IncidentUpdate represents a single update to an incident.
    One incident can have many incident updates.
    Only updated fields are stored in the incident update.
    """

    objects: IncidentUpdateManager = IncidentUpdateManager()
    # TODO Separate all the statuses (now milestones)

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    message = models.TextField(blank=True, null=True)
    _status = models.IntegerField(
        db_column="status",
        choices=IncidentStatus.choices,
        null=True,
        blank=True,
        verbose_name="Status",
    )
    severity = models.ForeignKey(
        Severity,
        null=True,
        blank=True,
        on_delete=models.SET(Severity.get_default),
        help_text="Superseded by priority",
    )
    severity.system_check_deprecated_details = {
        "msg": "The IncidentUpdate.severity field has been deprecated.",
        "hint": "Use IncidentUpdate.priority instead.",
        "id": "fields.W920",
    }
    priority = models.ForeignKey(
        Priority, null=True, blank=True, on_delete=models.SET(Priority.get_default)
    )
    environment = models.ForeignKey(
        Environment,
        null=True,
        blank=True,
        on_delete=models.SET(Environment.get_default),
    )
    incident = models.ForeignKey(
        "Incident", on_delete=models.CASCADE
    )
    component = models.ForeignKey(
        Component, null=True, blank=True, on_delete=models.SET_NULL
    )
    created_by = models.ForeignKey(
        User, null=True, blank=True, on_delete=models.SET_NULL
    )
    commander = models.ForeignKey(
        User,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="commander_for",
    )
    communication_lead = models.ForeignKey(
        User,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="communication_lead_for",
    )

    title = models.CharField(null=True, blank=True, max_length=128)
    description = models.TextField(
        null=True,
        blank=True,
    )

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    event_ts: datetime.datetime = models.DateTimeField(
        auto_now_add=False, editable=True
    )  # type: ignore
    event_type = models.CharField(null=True, blank=True, max_length=64)

    class Meta(TypedModelMeta):
        ordering = ["-event_ts"]

        constraints = [
            models.CheckConstraint(
                name="%(app_label)s_%(class)s__status_valid",
                check=models.Q(_status__in=IncidentStatus.values),
            )
        ]

    def __str__(self) -> str:
        if self.status:
            return f"{self.status.label}: {self.event_ts}"
        return f"{self.event_ts}"

    @property
    def status(self) -> IncidentStatus | None:
        if self._status:
            return IncidentStatus(self._status)
        return None

    @status.setter
    def status(self, status: IncidentStatus) -> None:
        self._status = status


@receiver(pre_save, sender=IncidentUpdate)
# pylint: disable=unused-argument
def set_event_ts(sender: Any, instance: IncidentUpdate, **kwargs: Any) -> None:
    """Add a timestamp on every IncidentUpdate.
    Not implemented using Django ORM auto_add as it would not make them user editable.
    """
    if instance.event_ts is None:
        instance.event_ts = timezone.now()  # type: ignore



from __future__ import annotations

import logging
from datetime import UTC, datetime
from typing import Any

from firefighter.slack.models.message import Message
from firefighter.slack.slack_app import SlackApp

app = SlackApp()

logger = logging.getLogger(__name__)


@app.event(
    {"type": "message", "subtype": "message_deleted"},
)
def handle_message_deleted(event: dict[str, Any]) -> None:
    channel_id = event.get("channel")
    ts = event.get("deleted_ts")
    if not channel_id or not ts:
        raise ValueError("Missing channel_id or ts")
    ts_dt = datetime.fromtimestamp(float(ts), tz=UTC)

    # Delete from DB if exists
    Message.objects.filter(
        conversation__channel_id=channel_id,
        ts=ts_dt,
    ).delete()



from __future__ import annotations

import pytest

from firefighter.slack.models.conversation import Conversation
from tests.test_slack.conftest import MockWebClient


@pytest.mark.django_db
def test_update_topic(
    conversation: Conversation, mock_web_client: MockWebClient
) -> None:
    mock_web_client.conversations_setTopic.return_value = {"ok": True}

    conversation.update_topic("New Topic", client=mock_web_client)
    assert mock_web_client.conversations_setTopic.called


@pytest.mark.django_db
def test_conversations_join(
    conversation: Conversation, mock_web_client: MockWebClient
) -> None:
    mock_web_client.conversations_join.return_value = {"ok": True}

    conversation.conversations_join(client=mock_web_client)
    assert mock_web_client.conversations_join.called



from __future__ import annotations

import logging
from typing import Any

from django.apps import apps
from django.db.models.query import Prefetch

from firefighter.firefighter.views import CustomDetailView
from firefighter.incidents.models.user import User

logger = logging.getLogger(__name__)
SELECT_RELATED = []
if apps.is_installed("firefighter.slack"):
    SELECT_RELATED.append("slack_user")
    from firefighter.slack.models import Conversation, UserGroup
if apps.is_installed("firefighter.pagerduty"):
    SELECT_RELATED.append("pagerduty_user")


class UserDetailView(CustomDetailView[User]):
    """In this view, be extra careful.

    In this context, `user` is the logged_in user and `target_user` is the user of the profile being viewed.
    """

    template_name = "pages/user_detail.html"
    context_object_name: str = "target_user"
    pk_url_kwarg = "user_id"
    model = User
    select_related = SELECT_RELATED
    queryset = User.objects.select_related(*select_related).prefetch_related(
        Prefetch(
            "conversation_set",
            queryset=Conversation.objects.not_incident_channel()
            .exclude(tag="")
            .filter(components__isnull=False)
            .distinct()
            .prefetch_related(Prefetch("members")),
        ),
        Prefetch(
            "usergroup_set",
            queryset=UserGroup.objects.filter(components__isnull=False)
            .distinct()
            .prefetch_related(Prefetch("members")),
        ),
    )

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        target_user: User = context[self.context_object_name]

        additional_context = {
            "page_title": f"{target_user.full_name} | User",
        }

        return {**context, **additional_context}



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django.dispatch.dispatcher import receiver

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.signals import incident_updated
from firefighter.raid.client import client

if TYPE_CHECKING:
    from firefighter.incidents.models.incident import Incident
    from firefighter.incidents.models.incident_update import IncidentUpdate

logger = logging.getLogger(__name__)


@receiver(signal=incident_updated, sender="update_status")
def incident_updated_close_ticket_when_mitigated_or_postmortem(
    sender: Any,
    incident: Incident,
    incident_update: IncidentUpdate,
    updated_fields: list[str],
    **kwargs: Any,
) -> None:
    # Close Jira ticket if mitigated or postmortem
    if "_status" in updated_fields and incident_update.status in {
        IncidentStatus.FIXED,
        IncidentStatus.POST_MORTEM,
    }:
        if not hasattr(incident, "jira_ticket") or incident.jira_ticket is None:
            logger.warning(
                f"Trying to close Jira ticket for incident {incident.id} but not having"
            )
            return
        client.close_issue(issue_id=incident.jira_ticket.id)
        # XXX We may want to add a comment if there is an incident update message on close



from __future__ import annotations

import pytest
from django.test import Client
from django.urls import reverse

from firefighter.incidents.models.user import User


@pytest.mark.django_db
def test_main_page(client: Client, footer_text: str, admin_user: User) -> None:
    """This test ensures that main page works."""
    client.force_login(admin_user)
    response = client.get("/")

    assert response.status_code == 200
    assert footer_text in str(response.content)


@pytest.mark.django_db
def test_hello_page(client: Client, footer_text: str, admin_user: User) -> None:
    """This test ensures that dashboard page works."""
    client.force_login(admin_user)
    response = client.get(reverse("incidents:dashboard"))

    assert response.status_code == 200
    assert footer_text in str(response.content)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Never

from django.dispatch.dispatcher import receiver

from firefighter.incidents import signals
from firefighter.pagerduty.models import PagerDutyUser

if TYPE_CHECKING:
    from firefighter.incidents.models import Incident
    from firefighter.incidents.models.user import User
logger = logging.getLogger(__name__)


@receiver(signal=signals.get_invites)
def get_invites_from_pagerduty(
    incident: Incident,
    **kwargs: Never,
) -> list[User]:
    if incident.private:
        return []
    return PagerDutyUser.objects.get_current_on_call_users_l1()



from __future__ import annotations

import logging
from functools import cache
from typing import Any, Generic, Never, TypeVar

from django.core.exceptions import BadRequest, PermissionDenied
from django.db import connection
from django.db.models import Model
from django.db.utils import OperationalError
from django.http import Http404, HttpRequest, JsonResponse
from django.urls import NoReverseMatch, reverse
from django.views.decorators.http import require_safe
from django.views.generic import DetailView

logger = logging.getLogger(__name__)

_MT = TypeVar("_MT", bound=Model)


@require_safe
def healthcheck(request: HttpRequest) -> JsonResponse:
    db_ok = True
    try:
        connection.ensure_connection()
    except OperationalError:
        logger.exception("Healthcheck failed: can't access DB!")
        db_ok = False

    if db_ok:
        return JsonResponse({"status": "ok"})
    return JsonResponse({"status": "fail"}, status=503)


def permission_denied_view(request: HttpRequest) -> Never:
    raise PermissionDenied("Permission denied")


def not_found_view(request: HttpRequest) -> Never:
    raise Http404("Not found")


def bad_request_view(request: HttpRequest) -> Never:
    raise BadRequest("Unknown parameters")


def server_error_view(request: HttpRequest) -> Never:
    # pylint: disable=broad-exception-raised
    raise Exception("Test exception for 500")  # noqa: TRY002


class CustomDetailView(DetailView[_MT], Generic[_MT]):
    """A custom detail view that adds the admin edit URL to the context, as `admin_edit_url`."""

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        obj: _MT = context[self.get_context_object_name(self.object) or "object"]
        return {**context, "admin_edit_url": get_admin_edit_url(self.model, obj.pk)}


@cache
def get_admin_edit_url(model_class: type[Model], object_pk: Any) -> str | None:
    """Construct the URL name for the admin edit page using the model's app_label and model_name."""
    app_label: str = model_class._meta.app_label  # noqa: SLF001
    model_name = model_class._meta.model_name  # noqa: SLF001
    url_name: str = f"admin:{app_label}_{model_name}_change"

    try:
        return reverse(url_name, args=[object_pk])
    except NoReverseMatch:
        return None



from __future__ import annotations

from typing import TYPE_CHECKING

from firefighter.slack.views.modals.close import CloseModal, modal_close
from firefighter.slack.views.modals.downgrade_workflow import (
    DowngradeWorkflowModal,
    modal_dowgrade_workflow,
)
from firefighter.slack.views.modals.edit import EditMetaModal, modal_edit
from firefighter.slack.views.modals.key_event_message import (  # XXX(dugab) move and rename (not a modal but a surface...)
    KeyEvents,
)
from firefighter.slack.views.modals.open import OpenModal, modal_open
from firefighter.slack.views.modals.opening import select_impact
from firefighter.slack.views.modals.opening.check_current_incidents import (
    CheckCurrentIncidentsModal,
)
from firefighter.slack.views.modals.opening.select_impact import modal_select_impact
from firefighter.slack.views.modals.postmortem import PostMortemModal, modal_postmortem
from firefighter.slack.views.modals.select import modal_select
from firefighter.slack.views.modals.send_sos import SendSosModal, modal_send_sos
from firefighter.slack.views.modals.status import StatusModal, modal_status
from firefighter.slack.views.modals.trigger_oncall import (
    OnCallModal,
    modal_trigger_oncall,
)
from firefighter.slack.views.modals.update import UpdateModal, modal_update
from firefighter.slack.views.modals.update_roles import (
    UpdateRolesModal,
    modal_update_roles,
)
from firefighter.slack.views.modals.update_status import (
    UpdateStatusModal,
    modal_update_status,
)

if TYPE_CHECKING:
    from firefighter.slack.views.modals.base_modal.base import SlackModal

selectable_modals: list[type[SlackModal]] = [
    UpdateModal,
    EditMetaModal,
    UpdateRolesModal,
    OnCallModal,
    CloseModal,
    PostMortemModal,
    StatusModal,
    SendSosModal,
    DowngradeWorkflowModal,
]



from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING

from django.core.validators import MinValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _
from django_stubs_ext.db.models import TypedModelMeta

from firefighter.incidents.models.milestone_type import (
    MilestoneType,
)


class MetricType(models.Model):
    name = models.CharField(
        max_length=50, help_text="Title or name (e.g. Time to Detect)"
    )
    code = models.CharField(
        max_length=10, unique=True, help_text="Short code (e.g. TTD)"
    )
    description = models.TextField()
    milestone_lhs = models.ForeignKey(
        MilestoneType,
        on_delete=models.CASCADE,
        related_name="metric_type_lhs",
        help_text="The milestone that ends the metric (duration = milestone_lhs - milestone_rhs)",
    )
    milestone_rhs = models.ForeignKey(
        MilestoneType,
        on_delete=models.CASCADE,
        related_name="metric_type_rhs",
        help_text="The milestone that starts the metric (duration = milestone_lhs - milestone_rhs)",
    )
    type = models.SlugField(
        max_length=50,
        unique=True,
        blank=True,
        help_text="Technical name, slugified version of the title (e.g. time_to_detect)",
    )

    class Meta(TypedModelMeta):
        constraints = [
            models.CheckConstraint(
                name="%(app_label)s_%(class)s__milestone_lhs_not_equal_milestone_rhs",
                check=~models.Q(milestone_lhs=models.F("milestone_rhs")),
                violation_error_message="The milestone that ends the metric cannot be the same as the milestone that starts the metric.",
            ),
        ]

        verbose_name = _("Metric Type")
        verbose_name_plural = _("Metric Types")

    def __str__(self) -> str:
        return self.name


class IncidentMetric(models.Model):
    incident = models.ForeignKey(
        "incidents.Incident",
        on_delete=models.CASCADE,
        db_index=True,
        related_name="metric_set",
    )
    metric_type = models.ForeignKey(MetricType, on_delete=models.CASCADE, db_index=True)
    duration = models.DurationField(
        blank=False, validators=[MinValueValidator(timedelta(seconds=0))]
    )

    if TYPE_CHECKING:
        incident_id: int

    class Meta(TypedModelMeta):
        verbose_name = _("Incident Metric")
        verbose_name_plural = _("Incident Metrics")
        constraints = [
            models.UniqueConstraint(
                name="%(app_label)s_%(class)s__incident_metric_type_unique",
                fields=["incident", "metric_type"],
            ),
            models.CheckConstraint(
                name="%(app_label)s_%(class)s__duration_positive",
                check=models.Q(duration__gte=timedelta(seconds=0)),
            ),
        ]

    def __str__(self) -> str:
        return f"#{self.incident_id} - {self.metric_type} - {self.duration}"



from __future__ import annotations

import logging
from typing import Any

from django.db.models.query import Prefetch

from firefighter.firefighter.views import CustomDetailView
from firefighter.incidents.models import Component

logger = logging.getLogger(__name__)


class ComponentDetailView(CustomDetailView[Component]):
    template_name = "pages/component_detail.html"
    context_object_name = "component"
    pk_url_kwarg = "component_id"
    model = Component
    select_related = ["group"]

    queryset = Component.objects.select_related(*select_related).prefetch_related(
        Prefetch("usergroups")
    )

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        component: Component = context["component"]

        additional_context = {
            "page_title": f"{component.name} Component",
        }

        return {**context, **additional_context}



from __future__ import annotations

import uuid

from django.db import models
from django.db.models.manager import Manager


class Group(models.Model):
    """Group of [firefighter.incidents.models.component.Component]. Not a group of users."""
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=128, unique=True)
    description = models.TextField(blank=True)
    order = models.IntegerField(default=0, unique=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    objects = Manager["Group"]()

    def __str__(self) -> str:
        return self.name



# Generated by Django 4.2.11 on 2024-05-03 09:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("raid", "0001_initial_oss"),
    ]

    operations = [
        migrations.CreateModel(
            name="FeatureTeam",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=80)),
                ("jira_project_key", models.CharField(max_length=10, unique=True)),
            ],
            options={
                "verbose_name": "Feature Team",
                "verbose_name_plural": "Feature Teams",
                "unique_together": {("name", "jira_project_key")},
            },
        ),
        migrations.RemoveField(
            model_name="qualifierrotation",
            name="jira_user",
        ),
        migrations.DeleteModel(
            name="Qualifier",
        ),
        migrations.DeleteModel(
            name="QualifierRotation",
        ),
    ]



from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING

import pytest
from django.contrib.admin.sites import site
from django.http import HttpResponseRedirect
from django.urls import reverse

if TYPE_CHECKING:
    from django.test import Client

    from firefighter.incidents.models.user import User

logger = logging.getLogger(__name__)


@pytest.mark.django_db
def test_health_check(client: Client) -> None:
    """This test ensures that health check is accessible."""
    response = client.get("/api/v2/firefighter/monitoring/healthcheck")

    assert response.status_code == 200


@pytest.mark.django_db
def test_admin_unauthorized(client: Client) -> None:
    """This test ensures that admin panel requires auth."""
    response = client.get("/admin/")

    assert response.status_code == 302


@pytest.mark.django_db
@pytest.mark.usefixtures("_debug")
def test_admin_authorized(admin_client: Client, admin_user: User) -> None:
    """This test ensures that admin panel is accessible."""
    admin_client.force_login(admin_user)

    response = admin_client.get("/")

    assert response.status_code == 200


@pytest.mark.django_db
def test_admin_docs_unauthorized(client: Client) -> None:
    """This test ensures that admin panel docs requires auth."""
    response = client.get("/admin/doc/")

    assert isinstance(response, HttpResponseRedirect)
    assert response.status_code == 302
    assert response.url == "/admin/login/?next=/admin/doc/"


@pytest.mark.django_db
def test_admin_docs_authorized(admin_client: Client) -> None:
    """This test ensures that admin panel docs are accessible."""
    response = admin_client.get("/admin/doc/")

    assert response.status_code == 200
    assert b"docutils" not in response.content


@pytest.mark.django_db
def test_robotstxt_present(client: Client) -> None:
    """This test ensures that robots.txt is present."""
    response = client.get("/robots.txt")

    assert response.status_code == 200
    assert b"Disallow: /" in response.content


@pytest.mark.django_db
def test_login_sso_button_admin_login(client: Client) -> None:
    """This test ensures that login page contains the SSO button."""
    response = client.get("/admin/login/")

    assert response.status_code == 200
    assert b"Log in with SSO" in response.content


@pytest.mark.django_db
def test_error_500_pretty(client: Client) -> None:
    """This test ensures that 500 error page is pretty."""
    with pytest.raises(Exception, match="Test exception for 500"):  # noqa: PT012
        response = client.get("/err/500/")

        assert response.status_code == 500
        assert b"Please try again later, and report it" in response.content


@pytest.mark.django_db
def test_error_404_pretty(client: Client) -> None:
    """This test ensures that 404 error page is pretty."""
    response = client.get("/err/404/")

    assert response.status_code == 404
    assert b"Please check the URL in the address bar and try again" in response.content


@pytest.mark.django_db
def test_error_403_pretty(client: Client) -> None:
    """This test ensures that 403 error page is pretty."""
    response = client.get("/err/403/")

    assert response.status_code == 403
    assert b"You do not have permission to access this page." in response.content


@pytest.mark.django_db
def test_error_400_pretty(client: Client) -> None:
    """This test ensures that 400 error page is pretty."""
    response = client.get("/err/400/")

    assert response.status_code == 400
    assert b"Bad request" in response.content


def test_errors_json(client: Client) -> None:
    """This test ensures that error pages are in JSON when Accept header is set."""
    client.raise_request_exception = False
    for error_code in [400, 403, 404, 500]:
        response = client.get(f"/err/{error_code}/", {}, HTTP_ACCEPT="application/json")

        assert response.status_code == error_code
        assert response.headers["Content-Type"] == "application/json"
        assert json.loads(response.content).get("error") is not None


@pytest.mark.django_db
@pytest.mark.timeout(30)
@pytest.mark.usefixtures("_debug")
def test_all_admin_list_no_error(admin_client: Client, admin_user: User) -> None:
    """This test ensures that admin panel is accessible."""
    admin_client.force_login(admin_user)

    # Get all list admin urls for models that have an admin
    urls = [
        reverse(f"admin:{model._meta.app_label}_{model._meta.model_name}_changelist")
        for model, model_admin in site._registry.items()
        if model_admin
    ]

    # Iterate over the list of urls and use the client to make get requests
    for url in urls:
        response = admin_client.get(url)
        # Check the status code of the response
        assert response.status_code == 200, f"Failed url is: {url}"



# Generated by Django 4.2.7 on 2023-12-04 10:41

import uuid

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    replaces = [
        ("slack", "0001_initial_squashed_0022_alter_sos_user_group"),
        ("slack", "0023_delete_incidentchannel_incidentchannel"),
    ]

    dependencies = [
        ("incidents", "0026_alter_user_options"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        (
            "incidents",
            "0013_environment_value_squashed_0017_rename_slack_groups_component_slack_conversations",
        ),
        ("incidents", "0023_auto_20220622_1559"),
        ("incidents", "0022_alter_user_name"),
        ("incidents", "0011_auto_20210628_1427"),
        ("incidents", "0027_remove_component_slack_conversations"),
        ("incidents", "0014_auto_20210809_1052"),
    ]

    operations = [
        migrations.CreateModel(
            name="Conversation",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(blank=True, max_length=80)),
                ("channel_id", models.CharField(max_length=128)),
                (
                    "_status",
                    models.IntegerField(
                        choices=[(0, "Not Opened"), (1, "Opened"), (2, "Archived")],
                        db_column="status",
                        default=0,
                    ),
                ),
                (
                    "_type",
                    models.IntegerField(
                        choices=[
                            (0, "Unknown"),
                            (10, "Public Channel"),
                            (20, "Private Channel"),
                            (30, "DM"),
                            (40, "Group DM"),
                        ],
                        db_column="type",
                        default=0,
                    ),
                ),
                ("is_shared", models.BooleanField(default=False)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "tag",
                    models.CharField(
                        blank=True,
                        help_text="Used by FireFighter internally to mark special conversations (#it-deploy, #tech-incidents...). Must be empty or unique.",
                        max_length=80,
                    ),
                ),
                (
                    "components",
                    models.ManyToManyField(
                        blank=True,
                        related_name="conversations",
                        to="incidents.component",
                    ),
                ),
                (
                    "members",
                    models.ManyToManyField(blank=True, to=settings.AUTH_USER_MODEL),
                ),
            ],
        ),
        migrations.CreateModel(
            name="IncidentChannel",
            fields=[
                (
                    "conversation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="slack.conversation",
                    ),
                ),
            ],
            bases=("slack.conversation",),
        ),
        migrations.CreateModel(
            name="UserGroup",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        blank=True,
                        help_text="Friendly name of the group. Corresponds to the `name` field in the Slack API. Max 80 characters.",
                        max_length=80,
                    ),
                ),
                (
                    "handle",
                    models.CharField(
                        blank=True,
                        help_text="Indicates the value used to notify group members via a mention. The handle does not include the leading @. Corresponds to the `handle` field in the Slack API. Max 80 characters.",
                        max_length=80,
                    ),
                ),
                (
                    "description",
                    models.CharField(
                        blank=True,
                        help_text="A short description of the group. Corresponds to the `description` field in the Slack API. Max 140 characters.",
                        max_length=140,
                    ),
                ),
                (
                    "usergroup_id",
                    models.CharField(
                        blank=True,
                        help_text="The Slack usergroup ID of the usergroup. Usually starts with `S`.",
                        max_length=32,
                        null=True,
                        unique=True,
                    ),
                ),
                (
                    "is_external",
                    models.BooleanField(
                        default=False,
                        help_text="Is this an external group, from an external Slack Workspace? Corresponds to the `is_external` field in the Slack API.",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "components",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Incident created with this usergroup automatically add the group members to these components.",
                        related_name="usergroups",
                        to="incidents.component",
                    ),
                ),
                (
                    "members",
                    models.ManyToManyField(blank=True, to=settings.AUTH_USER_MODEL),
                ),
            ],
            options={
                "verbose_name": "Slack user group",
                "verbose_name_plural": "Slack user groups",
            },
        ),
        migrations.CreateModel(
            name="Sos",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                (
                    "conversation",
                    models.ForeignKey(
                        blank=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="slack.conversation",
                    ),
                ),
                (
                    "user_group",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="slack.usergroup",
                    ),
                ),
            ],
            options={
                "verbose_name": "SOS",
                "verbose_name_plural": "SOS",
            },
        ),
        migrations.CreateModel(
            name="SlackUser",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("slack_id", models.CharField(max_length=32, unique=True)),
                (
                    "username",
                    models.CharField(
                        blank=True,
                        help_text="The Slack handle of a Slack user (@handle)",
                        max_length=32,
                        null=True,
                    ),
                ),
                ("image", models.URLField(blank=True, null=True)),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="slack_user",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Message",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("ts", models.DateTimeField(help_text="UTC Timestamp of the message")),
                ("thread_ts", models.DateTimeField(blank=True, null=True)),
                ("type", models.CharField(max_length=80)),
                ("subtype", models.CharField(blank=True, max_length=80)),
                ("text", models.TextField(blank=True, max_length=40000)),
                ("blocks", models.JSONField(blank=True, null=True)),
                ("ff_type", models.CharField(blank=True, max_length=64)),
                (
                    "conversation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="slack.conversation",
                    ),
                ),
                (
                    "incident",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="slack_message",
                        to="incidents.incident",
                    ),
                ),
                (
                    "incident_update",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="slack_message",
                        to="incidents.incidentupdate",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="slack.slackuser",
                    ),
                ),
            ],
        ),
        migrations.AddIndex(
            model_name="slackuser",
            index=models.Index(
                fields=["slack_id"], name="slack_slack_slack_i_06073f_idx"
            ),
        ),
        migrations.AddConstraint(
            model_name="message",
            constraint=models.UniqueConstraint(
                fields=("ts", "conversation"), name="message_ts_unicity"
            ),
        ),
        migrations.AddField(
            model_name="incidentchannel",
            name="incident",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="conversation",
                to="incidents.incident",
            ),
        ),
        migrations.AddConstraint(
            model_name="conversation",
            constraint=models.CheckConstraint(
                check=models.Q(("_type__in", [0, 10, 20, 30, 40])),
                name="slack_conversation__type_valid",
            ),
        ),
        migrations.AddConstraint(
            model_name="conversation",
            constraint=models.CheckConstraint(
                check=models.Q(("_status__in", [0, 1, 2])),
                name="slack_conversation__status_valid",
            ),
        ),
        migrations.AddConstraint(
            model_name="conversation",
            constraint=models.UniqueConstraint(
                condition=models.Q(("tag", ""), _negated=True),
                fields=("tag",),
                name="unique__tag",
            ),
        ),
    ]



from __future__ import annotations

from firefighter.slack.models.conversation import Conversation
from firefighter.slack.models.incident_channel import IncidentChannel
from firefighter.slack.models.message import Message
from firefighter.slack.models.user import SlackUser
from firefighter.slack.models.user_group import UserGroup



from __future__ import annotations

import logging
from typing import Any

from firefighter.pagerduty.client import PagerdutyClient

logger = logging.getLogger(__name__)


class PagerdutyService:
    """XXX Rename to PagerDutyClient to avoid confusion with PagerDutyService Django model."""

    logger: logging.Logger = logging.getLogger(__name__)
    client: PagerdutyClient

    def __init__(self) -> None:
        self.client = PagerdutyClient()

    def get_phone_number_from_body(self, pd_user: dict[str, Any]) -> str | None:
        phone_number = None
        contact_methods: list[dict[str, str | int | None]] = pd_user.get(
            "contact_methods", []
        )
        # Prefer "Work" phone number, get anything we can
        for summary in ["Work", "Mobile", "SMS", "Phone", "Home"]:
            # Get the first dict element of the list of contact methods with the right summary
            best_phone = next(
                (
                    contact_method
                    for contact_method in contact_methods
                    if contact_method["summary"] == summary
                    and (
                        contact_method["type"]
                        in {
                            "phone_contact_method",
                            "phone_contact_method_reference",
                            "sms_contact_method",
                        }
                    )
                ),
                None,
            )
            if best_phone:
                if best_phone["type"] in {"phone_contact_method", "sms_contact_method"}:
                    return f"{best_phone['country_code']}{best_phone['address']}"

                if (
                    best_phone["type"] == "phone_contact_method_reference"
                    and best_phone["id"] is not None
                ):
                    phone_number = self.get_user_phone_number(
                        pd_user["id"], str(best_phone["id"])
                    )
                    if phone_number is not None and phone_number != "":
                        return phone_number

        return phone_number

    def get_user_phone_number(
        self, pagerduty_user_id: str, contact_method_id: str
    ) -> str:
        contact_method: dict[str, Any] = (
            self.client.get_user_contact_method(pagerduty_user_id, contact_method_id)
        ).json()["contact_method"]
        if contact_method["type"] != "phone_contact_method":
            err_msg = f"Wrong contact method, expecting phone_contact_method, got {contact_method['type']}"
            raise AttributeError(err_msg)
        return f"{contact_method['country_code']}{contact_method['address']}"

    def get_all_oncalls(self) -> list[dict[str, Any]]:
        return self.client.session.list_all(
            "oncalls",
            params={"include[]": ["users", "schedules", "escalation_policies"]},
        )


pagerduty_service = PagerdutyService()



from __future__ import annotations

import enum
import logging
import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Never

from slack_sdk.models.blocks.basic_components import MarkdownTextObject
from slack_sdk.models.blocks.block_elements import ButtonElement
from slack_sdk.models.blocks.blocks import Block, SectionBlock
from slack_sdk.models.metadata import Metadata

from firefighter.slack.slack_app import DefaultWebClient, slack_client

if TYPE_CHECKING:
    from slack_sdk.web.client import WebClient
    from slack_sdk.web.slack_response import SlackResponse

    from firefighter.incidents.models.incident import Incident
logger = logging.getLogger(__name__)
VALID_ID_REGEX = re.compile(r"^[a-z][a-z0-9_]*$")


class SlackMessageStrategy(enum.Enum):
    """Define how should the message be posted.

    - `append`: add the message to the channel, even if a message with the same type has already been posted.
    - `replace`: replace the last message with the same type (push new one, delete old one)
    - `update`: update in place the last message with the same type
    """

    APPEND = "append"
    REPLACE = "replace"
    UPDATE = "update"


class SlackMessageSurface(ABC):
    """Base class for Slack messages, which are sent to a channel or a user.

    This provides a common interface to send messages with text, blocks and metadata.

    This is helpful to send messages but also to save them in DB.
    """

    id: str
    strategy: SlackMessageStrategy = SlackMessageStrategy.APPEND
    """Alphanumeric ID, starting with a letter, that may contain underscores. """

    def __init__(self) -> None:
        if not self.id or self.id == "":
            logger.warning(f"No Slack message ID set for {self.__class__.__name__}.")
        # Check the ID is well-formed to be a Slack event_type: alphanumeric string, starting with a letter, containing underscore
        elif not VALID_ID_REGEX.match(self.id):
            logger.warning(
                f"Slack message ID {self.id} is not well formed for {self.__class__.__name__}."
            )

    def get_blocks(self) -> list[Block]:
        """Returns the blocks of the message.

        Returns:
            list[Block]: List of Slack Blocks for the message. Default is the text as a SectionBlock.
        """
        return [SectionBlock(text=self.get_text())]

    @abstractmethod
    def get_text(self) -> str:
        """Returns the text of the message."""
        raise NotImplementedError

    def get_metadata(self) -> Metadata:
        """The value of `event_type` should be an alphanumeric string, and human-readable. The value of this field may appear in the UI to developers, so keep this in mind when choosing a value. Developers should make an effort to name with the pattern <resource_name_singular> and <action_in_past_tense>
        See more https://api.slack.com/reference/metadata.

        Returns:
            Metadata: Slack Metadata for the message.
        """
        return Metadata(event_type=self.id, event_payload={"ff_type": self.id})

    def get_slack_message_options(self) -> dict[str, Any]:
        return {
            "unfurl_links": False,
        }

    def get_slack_message_params(
        self, *, blocks_as_dict: bool = False
    ) -> dict[str, Any]:
        blocks: list[Block] | list[dict[str, Any]]
        blocks = (
            [block.to_dict() for block in self.get_blocks()]
            if blocks_as_dict
            else self.get_blocks()
        )

        return {
            "blocks": blocks,
            "text": self.get_text(),
            "metadata": self.get_metadata().to_dict(),
            **self.get_slack_message_options(),
        }

    @slack_client
    def post_message(
        self,
        conversation_id: str,
        client: WebClient = DefaultWebClient,
        **kwargs: Never,
    ) -> SlackResponse:
        """Deprecated. Only use for Global Channel, until the global channel is set in DB."""
        return client.chat_postMessage(
            channel=conversation_id,
            **self.get_slack_message_params(),
        )


class SectionBlockUpdateIntent(SectionBlock):
    def __init__(self, incident: Incident, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.text = MarkdownTextObject(
            text="_Make an update with `/incident update` or this button :point_right:_"
        )

        self.accessory = ButtonElement(
            text="Update",
            action_id="open_modal_incident_update_status",
            value=str(incident.id),
        )



"""This module define all the settings for the FireFighter project.

!!! warning

    **It should NOT be imported directly from any other module.**

    To use the settings, import them through Django's [django.conf.settings](https://docs.djangoproject.com/en/4.2/topics/settings/#using-settings-in-python-code) object:
    ```python
    from django.conf import settings
    ```

"""

# ruff: noqa: E402, F403
# pylint: disable=wrong-import-position
# isort: off
from __future__ import annotations

import os

os.environ["DJANGO_SETTINGS_MODULE"] = "firefighter.firefighter.settings"


# Monkey-patching Django, so stubs will work for all generics,
# see: https://github.com/typeddjango/django-stubs
import django_stubs_ext

django_stubs_ext.monkeypatch()

# Allow mapping of settings to environment variables
from firefighter.firefighter.settings.settings_utils import config
from decouple import Csv

FF_ENV_VAR_MAPPING_LIST = config("FF_ENV_VAR_MAPPING", cast=Csv(), default="")
FF_ENV_VAR_MAPPING = {
    k: v for k, v in (x.split(":") for x in FF_ENV_VAR_MAPPING_LIST) if k and v
}
for k, v in FF_ENV_VAR_MAPPING.items():
    os.environ[k] = os.environ.get(v, "")

from firefighter.firefighter.settings.settings_builder import *

# Monkey-patching DRF, to allow generics on some classes
# We need to do this after importing settings, as DRF uses them at import time
from rest_framework import fields, generics, viewsets

django_stubs_ext.monkeypatch(
    extra_classes=(fields.Field, generics.GenericAPIView, viewsets.GenericViewSet)
)
# XXX(dugab): remove in Django50
from django.forms.renderers import BaseRenderer

BaseRenderer.form_template_name = "django/forms/div.html"
BaseRenderer.formset_template_name = "django/forms/formsets/div.html"



"""WSGI config for firefighter project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/4.2/howto/deployment/wsgi/
"""

from __future__ import annotations

# We need to be able to shadow virtualenv things with our own code
# So we reverse the order of the last two parts of the load path
import os

# Load the WSGI Application (as we should do)


os.environ.setdefault("DJANGO_SETTINGS_MODULE", "firefighter.firefighter.settings")
# pylint: disable=wrong-import-position
# noinspection PyPep8
from django.core.wsgi import get_wsgi_application

application = get_wsgi_application()



from __future__ import annotations

from importlib import import_module

# pylint: disable=wildcard-import
# isort: off
from firefighter.firefighter.settings.settings_utils import SETTINGS_ENV, config
from firefighter.firefighter.settings.components.api import *
from firefighter.firefighter.settings.components.caches import *
from firefighter.firefighter.settings.components.celery import *
from firefighter.firefighter.settings.components.common import *
from firefighter.firefighter.settings.components.confluence import *
from firefighter.firefighter.settings.components.jira_app import *
from firefighter.firefighter.settings.components.logging import *
from firefighter.firefighter.settings.components.pagerduty import *
from firefighter.firefighter.settings.components.slack import *
from firefighter.firefighter.settings.components.raid import *

# Load dev or prod settings:
if SETTINGS_ENV == "dev":
    from firefighter.firefighter.settings.environments.dev import *
elif SETTINGS_ENV == "prod":
    from firefighter.firefighter.settings.environments.prod import *
else:
    err_msg = f"Unknown environment: {SETTINGS_ENV}"
    raise ValueError(err_msg)


FF_ADDITIONAL_SETTINGS_MODULE = config("FF_ADDITIONAL_SETTINGS_MODULE", default=None)

if FF_ADDITIONAL_SETTINGS_MODULE:
    try:
        m = import_module(FF_ADDITIONAL_SETTINGS_MODULE)
        # Set locals from the module
        for item in dir(m):
            if not item.startswith("__"):
                globals()[item] = getattr(m, item)

    except ImportError as exc:
        msg = f"Couldn't import {FF_ADDITIONAL_SETTINGS_MODULE}. Are you sure it's installed and available on your PYTHONPATH environment variable? Did you forget to activate a virtual environment?"
        raise ImportError(msg) from exc



from __future__ import annotations

from django.apps import apps
from django.urls import path
from rest_framework import routers

from firefighter.api.urls import urlpatterns as api_urlspatterns
from firefighter.confluence.views.api import RunbookViewSet
from firefighter.confluence.views.runbook.runbook_list import RunbooksViewList

app_name = "confluence"
urlpatterns = [
    path("runbook/", RunbooksViewList.as_view(), name="runbook_list"),
]


# Patch `api` app router, if installed
if apps.is_installed("firefighter.api"):
    router = routers.DefaultRouter()
    router.include_root_view = False
    router.register(r"runbooks", RunbookViewSet, basename="runbooks")
    api_urlspatterns.extend(router.urls)



from __future__ import annotations

from django.db import models


class IncidentStatus(models.IntegerChoices):
    OPEN = 10, "Open"
    INVESTIGATING = 20, "Investigating"
    FIXING = 30, "Mitigating"
    FIXED = 40, "Mitigated"
    POST_MORTEM = 50, "Post-mortem"
    CLOSED = 60, "Closed"

    @staticmethod
    def lt(val: int) -> list[IncidentStatus]:
        return [i for i in IncidentStatus if i.value < val]

    @staticmethod
    def lte(val: int) -> list[IncidentStatus]:
        return [i for i in IncidentStatus if i.value <= val]

    @staticmethod
    def gt(val: int) -> list[IncidentStatus]:
        return [i for i in IncidentStatus if i.value > val]

    @staticmethod
    def gte(val: int) -> list[IncidentStatus]:
        return [i for i in IncidentStatus if i.value >= val]

    @staticmethod
    def choices_lt(val: int) -> list[tuple[int, str]]:
        return [i for i in IncidentStatus.choices if i[0] < val]



from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING

from django.conf import settings
from slack_sdk.models.blocks.basic_components import MarkdownTextObject
from slack_sdk.models.blocks.blocks import (
    Block,
    DividerBlock,
    SectionBlock,
)

from firefighter.slack.messages.base import SlackMessageSurface
from firefighter.slack.slack_templating import user_slack_handle_or_name

RAID_JIRA_API_URL: str = settings.RAID_JIRA_API_URL

if TYPE_CHECKING:
    from firefighter.incidents.models.user import User
    from firefighter.raid.models import JiraTicket


class SlackMessageRaidCreatedIssue(SlackMessageSurface):
    id = "raid_created_issue"

    def __init__(self, ticket: JiraTicket, reporter_user: User | None = None) -> None:
        self.ticket = ticket
        self.reporter_user = reporter_user or ticket.reporter.user
        super().__init__()

    def get_text(self) -> str:
        return f"New {self.ticket.issue_type}: {self.ticket.key} '{self.ticket.summary}' was created."

    def get_blocks(self) -> list[Block]:
        blocks: list[Block] = [
            SectionBlock(
                text=f"New {self.ticket.issue_type}: *{self.ticket.key}* _ {self.ticket.summary} _ was created."
            ),
            # XXX Add issue type emoji from RAID
            DividerBlock(),
            SectionBlock(
                fields=[
                    MarkdownTextObject(text=f":tag: *Title:*\n{self.ticket.summary}"),
                    MarkdownTextObject(
                        text=textwrap.shorten(
                            f":spiral_note_pad: *Description:*\n{self.ticket.description}",
                            width=1997,
                            placeholder="...",
                        )
                    ),
                    MarkdownTextObject(
                        text=f":happydoge: *Reporter:*\n{user_slack_handle_or_name(self.reporter_user)}"
                    ),
                    MarkdownTextObject(text=f":jira_new: *Jira:*\n{self.ticket.url}"),
                    MarkdownTextObject(
                        text=f":warning: *Business impact*: {self.ticket.business_impact or 'N/A'}"
                    ),
                ]
            ),
        ]
        return blocks


class SlackMessageRaidModifiedIssue(SlackMessageSurface):
    id = "raid_modify_issue"

    def __init__(
        self,
        jira_ticket_key: str,
        jira_author_name: str,
        jira_field_modified: str,
        jira_field_from: str,
        jira_field_to: str,
    ) -> None:
        self.jira_ticket_key = jira_ticket_key
        self.jira_author_name = jira_author_name
        self.jira_field_modified = jira_field_modified
        self.jira_field_from = jira_field_from
        self.jira_field_to = jira_field_to
        super().__init__()

    def get_text(self) -> str:
        return f"Modified Jira ticket {self.jira_ticket_key}"

    def get_blocks(self) -> list[Block]:
        return [
            SectionBlock(
                text=f":jira_new: Jira ticket *{self.jira_ticket_key}* was *modified* :hammer:"
            ),
            DividerBlock(),
            SectionBlock(
                fields=[
                    MarkdownTextObject(
                        text=f":face_with_monocle: *Author:*\n{self.jira_author_name}"
                    ),
                    MarkdownTextObject(
                        text=f":tag: *Field:*\n{self.jira_field_modified}"
                    ),
                    MarkdownTextObject(text=f":x: *From:*\n{self.jira_field_from}"),
                    MarkdownTextObject(
                        text=f":light_check_mark: *To:*\n{self.jira_field_to}"
                    ),
                    MarkdownTextObject(
                        text=f":jira_new: *URL:*\n{RAID_JIRA_API_URL}/browse/{self.jira_ticket_key}"
                    ),
                ]
            ),
        ]


class SlackMessageRaidComment(SlackMessageSurface):
    id = "raid_comment"

    def __init__(
        self,
        jira_ticket_key: str,
        author_jira_name: str,
        comment: str,
        webhook_event: str,
    ) -> None:
        self.jira_ticket_key = jira_ticket_key
        self.author_jira_name = author_jira_name
        self.comment = comment
        self.webhook_event = webhook_event
        match self.webhook_event:
            case "comment_created":
                self.comment_event_text = ":green_book: New comment added"
            case "comment_updated":
                self.comment_event_text = ":blue_book: Updated comment"
            case "comment_deleted":
                self.comment_event_text = ":closed_book: Deleted comment"
        super().__init__()

    def get_text(self) -> str:
        return f"{self.comment_event_text} on {self.jira_ticket_key}"

    def get_blocks(self) -> list[Block]:
        return [
            SectionBlock(
                text=f"{self.comment_event_text} on the Jira ticket *{self.jira_ticket_key}*"
            ),
            DividerBlock(),
            SectionBlock(
                fields=[
                    MarkdownTextObject(
                        text=f":face_with_monocle: *Author:*\n{self.author_jira_name}"
                    ),
                    MarkdownTextObject(text=f":pencil2: *Comment:*\n{self.comment}"),
                    MarkdownTextObject(
                        text=f":jira_new: *Jira:*\n{RAID_JIRA_API_URL}/browse/{self.jira_ticket_key}"
                    ),
                ]
            ),
        ]



from __future__ import annotations

import logging
import time
import urllib.parse
from functools import cached_property
from typing import Any, cast

from django import db
from django.conf import settings
from jira import JIRA, exceptions
from jira import User as JiraAPIUser

from firefighter.incidents.models.user import User
from firefighter.jira_app.models import JiraUser
from firefighter.jira_app.types import (
    Status,
    StatusTransitionInfo,
    Transition,
    WorkflowBuilderResponse,
)
from firefighter.jira_app.utils import (
    get_status_id_from_name,
    get_transitions_to_apply,
    pythonic_keys,
)

logger = logging.getLogger(__name__)


RAID_JIRA_API_URL: str = settings.RAID_JIRA_API_URL


class JiraUserNotFoundError(Exception):
    pass


class JiraUserDatabaseError(Exception):
    pass


class JiraAPIError(Exception):
    pass


class SlackNotificationError(Exception):
    pass


class GetWatchersError(Exception):
    pass


class GetReporterError(Exception):
    pass


class JiraClient:
    def __init__(self) -> None:
        self.url = RAID_JIRA_API_URL

    @cached_property
    def jira(self) -> JIRA:
        return JIRA(
            server=self.url,
            basic_auth=(
                settings.RAID_JIRA_API_USER,
                settings.RAID_JIRA_API_PASSWORD,
            ),
            options={"headers": settings.FF_HTTP_CLIENT_ADDITIONAL_HEADERS or {}},
        )

    def transition_issue_auto(
        self, issue_id: str | int, target_status_name: str, workflow_name: str
    ) -> None:
        """Attempts to close an issue by applying transitions to it.

        Args:
            issue_id (str | int): Jira issue id
            target_status_name (str): target status name
            workflow_name (str): workflow name
        """
        issue_id = str(issue_id)
        transitions_info = self._get_transitions(
            self._get_project_config_workflow_from_builder_base(workflow_name)
        )
        if len(transitions_info) == 0:
            logger.error(
                f"Could not find transitions for issue id={issue_id}! Not closing issue."
            )

        # Get closed state id
        # XXX Use a list of closed states to support multiple workflows, or better
        closed_state_id = get_status_id_from_name(transitions_info, target_status_name)
        if closed_state_id is None:
            logger.warning(
                f"Could not find target status '{target_status_name}' id for issue {issue_id}! Not closing issue."
            )
            return

        # Get current issue status
        issue = self.jira.issue(issue_id)
        current_status_id = int(issue.fields.status.id)

        # Get transitions to apply
        transitions_to_apply = get_transitions_to_apply(
            current_status_id, transitions_info, closed_state_id
        )

        if len(transitions_to_apply) == 0:
            logger.info(f"Issue {issue_id} is already closed. Not closing again.")

        # Apply transitions
        # XXX Better error handling
        for transition in transitions_to_apply:
            logger.debug(f"Running transition: {transition}")
            self.jira.transition_issue(
                issue=issue_id,
                transition=transition,
                fields={},
            )

    def _fetch_jira_user(self, username: str) -> JiraAPIUser:
        """Fetches a Jira user from the Jira API.

        Args:
            username (str): username of the Jira user (e.g. john.doe)

        Raises:
            ValueError: Empty username
            JiraUserNotFoundError: User not found in Jira

        Returns:
            JiraAPIUser: Jira API user object
        """
        if username is None or username == "":
            invalid_username_log = f"Invalid username '{username}'"
            raise ValueError(invalid_username_log)
        response = self.jira.search_users(query=username)
        if len(response) == 0:
            raise JiraUserNotFoundError("User not found in Jira.")
        if len(response) > 1:
            # Get the first user that matches the email address
            jira_user = next(
                (
                    user
                    for user in response
                    if user.raw.get("emailAddress", "a@b").split("@")[0] == username
                ),
                None,
            )
            if jira_user is None:
                err_msg = f"More than one JIRA user found for username '{username}'."
                raise JiraUserNotFoundError(err_msg)
            return jira_user
        return response[0]

    def get_jira_user_from_user(self, user: User) -> JiraUser:
        """Fetches a Jira user from the Jira API.

        Args:
            user (User): User object

        Raises:
            JiraUserNotFoundError: User not found in Jira

        Returns:
            JiraAPIUser: Jira API user object
        """
        # Check if user has a Jira user
        if hasattr(user, "jira_user") and user.jira_user:
            return user.jira_user

        username = user.email.split("@")[0]
        jira_user = self._fetch_jira_user(username)

        return JiraUser.objects.update_or_create(
            id=jira_user.raw.get("accountId"), defaults={"user": user}
        )[0]

    def get_jira_user_from_jira_id(self, jira_account_id: str) -> JiraUser:
        """Look for a Jira User in DB, if not found, fetch it from Jira API.

        Args:
            jira_account_id (str): Jira account id

        Raises:
            JiraUserNotFoundError: User not found in Jira nor in DB
            JiraUserDatabaseError: Unable to create user in DB
            ValueError: Empty jira_account_id

        Returns:
            JiraUser: Jira user object
        """
        if jira_account_id is None or jira_account_id == "":
            err_msg = f"Jira account id is empty ('{jira_account_id}')"
            raise ValueError(err_msg)

        # Look in the DB
        try:
            return JiraUser.objects.get(id=jira_account_id)
        except JiraUser.DoesNotExist:
            logger.debug(
                f"Jira user {jira_account_id} not found in DB, fetching from Jira API"
            )
        logger.info("User %s not found in DB. Check sync user task.", jira_account_id)

        # Look on JIRA API
        jira_api_user, email = self._get_user_from_api(jira_account_id)

        username: str = email.split("@")[0]
        # Check if we have user with same email
        try:
            user: User = User.objects.select_related("jira_user").get(email=email)
            if (
                hasattr(user, "jira_user")
                and user.jira_user
                and isinstance(user.jira_user, JiraUser)
            ):
                return user.jira_user
            try:
                return JiraUser.objects.create(
                    id=jira_account_id,
                    user=user,
                )
            except db.IntegrityError as e:
                logger.exception("Error creating user %s", jira_account_id)
                raise JiraUserDatabaseError("Unable to create user") from e

        except User.DoesNotExist:
            logger.warning("User %s not found in DB. Creating it...", jira_account_id)
            user = self._create_user_from_jira_info(
                jira_account_id, jira_api_user, email, username
            )

        try:
            return JiraUser.objects.create(
                id=jira_account_id,
                user=user,
            )
        except db.IntegrityError as e:
            logger.exception("Error creating user %s", jira_account_id)
            raise JiraUserDatabaseError("Unable to create user") from e

    def get_watchers_from_jira_ticket(
        self, jira_issue_id: int | str
    ) -> list[JiraAPIUser]:
        """Fetch watchers for a specific Jira ticket from Jira API.

        Args:
            jira_issue_id (str | int): Jira issue id

        Raises:
            ValueError: Empty issue id

        Returns:
            list(JiraAPIUser): List of Jira users object
        """
        watchers = self.jira.watchers(jira_issue_id).raw.get("watchers")
        if len(watchers) == 0:
            logger.warning(
                "Watchers not found for jira_account_id '%s'.", jira_issue_id
            )
        return watchers

    @staticmethod
    def _create_user_from_jira_info(
        jira_account_id: str,
        jira_api_user: JiraAPIUser,
        email: str,
        username: str,
    ) -> User:
        name = jira_api_user.raw.get("displayName")
        if not name or not isinstance(name, str):
            logger.warning("User %s has no display name, using email as name", email)
            name = email.split("@")[0]
        try:
            user: User = User.objects.create(
                name=name,
                email=email,
                username=username,
            )

        except db.IntegrityError as e:
            logger.exception("Error creating user %s", jira_account_id)
            raise JiraUserDatabaseError("Unable to create user") from e
        return user

    def _get_user_from_api(self, jira_account_id: str) -> tuple[JiraAPIUser, str]:
        try:
            jira_api_user = self.jira.user(jira_account_id)
            email: str | None = jira_api_user.raw.get("emailAddress")

        except exceptions.JIRAError as e:
            logger.exception("Error getting user %s", jira_account_id)
            raise JiraUserNotFoundError("User not Found") from e

        if email is None:
            logger.warning("User %s has no email address", jira_account_id)
            raise JiraUserNotFoundError("User not Found: no email address for user")
        return jira_api_user, email

    def _get_project_config_workflow_base(
        self, project_key: str, workflow_name: str
    ) -> dict[str, Any]:
        """This is an undocumented API. Make sure your error handling is robust."""
        ts_ms = int(time.time() * 1000)
        # XXX Incident workflow name resolution
        # XXX Check if INCIDENT is configured the same way, try to get it dynamically
        workflow_name_url_encoded = urllib.parse.quote_plus(workflow_name)

        get_workflow = f"/rest/projectconfig/latest/workflow?projectKey={project_key}&workflowName={workflow_name_url_encoded}&_={ts_ms}"
        url = f"{self.jira.server_url}{get_workflow}"

        # XXX Try catch
        # XXX Override Jira Class / split client and service
        return cast(
            "dict[str, Any]",
            self.jira._session.get(  # noqa: SLF001
                url,
                headers=self.jira._options["headers"],  # noqa: SLF001
            ).json(),
        )

    def _get_project_config_workflow_from_builder_base(
        self, workflow_name: str
    ) -> WorkflowBuilderResponse:
        ts_ms = int(time.time() * 1000)
        # XXX Incident workflow name resolution
        # XXX Check if INCIDENT is configured the same way, try to get it dynamically
        workflow_name_url_encoded = urllib.parse.quote_plus(workflow_name)

        get_workflow = f"/rest/workflowDesigner/latest/workflows?name={workflow_name_url_encoded}&_={ts_ms}&draft=false"
        url = f"{self.jira.server_url}{get_workflow}"

        # XXX Try catch
        # XXX Override Jira Class / split client and service
        res = self.jira._session.get(  # noqa: SLF001
            url,
            headers=self.jira._options["headers"],  # noqa: SLF001
        ).json()
        statuses: list[dict[str, Any]] = res["layout"]["statuses"]
        transitions: list[dict[str, Any]] = res["layout"]["transitions"]
        statuses = pythonic_keys(statuses)
        transitions = pythonic_keys(transitions)
        for transition in transitions:
            # XXX: Ugly, make a helper and make sure it's safe
            transition["source_id"] = int(
                transition["source_id"]
                .replace("S<", "")
                .replace(">", "")
                .replace("I<", "")
            )
            transition["target_id"] = int(
                transition["target_id"]
                .replace("S<", "")
                .replace(">", "")
                .replace("I<", "")
            )
            if "source_angle" in transition:
                del transition["source_angle"]
            if "target_angle" in transition:
                del transition["target_angle"]

        for status in statuses:
            status["id"] = int(
                status["id"].replace("S<", "").replace(">", "").replace("I<", "")
            )
            if "x" in status:
                del status["x"]
            if "y" in status:
                del status["y"]

        return WorkflowBuilderResponse(
            statuses=cast("list[Status]", statuses),
            transitions=cast("list[Transition]", transitions),
        )

    @staticmethod
    def _get_transitions(
        workflow_builder_response: WorkflowBuilderResponse,
    ) -> list[StatusTransitionInfo]:
        statuses_info_list: list[StatusTransitionInfo] = []

        statuses: list[Status] = workflow_builder_response["statuses"]
        transitions: list[Transition] = workflow_builder_response["transitions"]

        step_id_to_status_id: dict[int, int] = {
            status["step_id"]: int(status.get("status_id", status["step_id"]))
            for status in statuses
        }
        status_id_to_name: dict[int, str] = {
            step_id_to_status_id[status["step_id"]]: status["name"]
            for status in statuses
        }

        for transition in transitions:
            if "screen_name" in transition:
                logger.debug(
                    "Ignoring Transition %s has a screen_name: %s",
                    transition["name"],
                    transition["screen_name"],
                )
                continue  # Ignore transitions that have a screen_name
            if transition["initial"]:
                logger.debug(f"Ignoring Transition {transition['name']} is initial")
                continue

            source_id = step_id_to_status_id[transition["source_id"]]
            target_id = step_id_to_status_id[transition["target_id"]]

            if (
                source_id not in step_id_to_status_id.values()
                or target_id not in step_id_to_status_id.values()
            ):
                logger.info(
                    f"Ignoring Transition {transition['name']} has invalid status: {source_id} and {target_id} not in {step_id_to_status_id}"
                )
                continue  # Ignore transitions that refer to non-existent statuses

            # Prepare _StatusTransitionInfo
            status_transition_info: StatusTransitionInfo = {
                "status_id": source_id,
                "status_name": status_id_to_name[source_id],
                "target_statuses": {target_id},
                "transition_to_status": {target_id: transition["name"]},
            }

            # Check if the status already exists in the list
            for existing_info in statuses_info_list:
                if existing_info["status_id"] == source_id:
                    existing_info["target_statuses"].add(target_id)
                    existing_info["transition_to_status"][target_id] = transition[
                        "name"
                    ]
                    break
            else:
                # If the status doesn't exist in the list, add the new info
                statuses_info_list.append(status_transition_info)

        # For global transitions, add the transition to all statuses
        for transition in transitions:
            if transition["global_transition"]:
                for status_info in statuses_info_list:
                    status_info["target_statuses"].add(
                        step_id_to_status_id[int(transition["target_id"])]
                    )
                    status_info["transition_to_status"][
                        step_id_to_status_id[int(transition["target_id"])]
                    ] = transition["name"]

        return statuses_info_list


client = JiraClient()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from firefighter.incidents.models.user import User

logger = logging.getLogger(__name__)


def link_auth_user(user: User, claim: dict[str, str | list[str]]) -> None:
    # ruff: noqa: PLC0415

    # Check that roles are in the claim
    group_names = claim.get("roles")
    if group_names is None or not isinstance(group_names, list):
        return

    # Map permissions

    from django.contrib.auth.models import Group

    # Check special "back_office" group which enables the connection to the BO
    if "back_office" in group_names:
        group_names.remove("back_office")
        if not user.is_staff:
            user.is_staff = True
            user.save()

    # Reset all groups
    user.groups.clear()
    # Add all groups received by SSO
    for group_name in group_names:
        try:
            group = Group.objects.get(name=group_name)
            group.user_set.add(
                user
            )  # pyright: ignore[reportGeneralTypeIssues]
        except Group.DoesNotExist:
            logger.warning(
                "Group %s from SSO does not exist in the application Groups!",
                group_name,
            )



from __future__ import annotations

import django_tables2 as tables

from firefighter.confluence.models import Runbook
from firefighter.firefighter.tables_utils import BASE_TABLE_ATTRS


class RunbookTable(tables.Table):
    class Meta:
        model = Runbook
        template_name = "incidents/table.html"
        fields = ("service_type",)
        attrs: dict[str, dict[str, str] | str] = {
            **BASE_TABLE_ATTRS,
            "aria-label": "Runbook list table",
        }

    # Disable for now as we link directly to Confluence
    name = tables.TemplateColumn(
        '<a href="{{record.page_url}}" class="text-primary hover:text-indigo-900 dark:text-indigo-400 dark:hover:text-indigo-200">{{record.name}}</a>'
    )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from slack_sdk.models.views import View

from firefighter.incidents.forms.edit import EditMetaForm
from firefighter.slack.views.modals.base_modal.base import ModalForm

if TYPE_CHECKING:
    from slack_bolt.context.ack.ack import Ack

    from firefighter.incidents.models.incident import Incident
    from firefighter.incidents.models.user import User
    from firefighter.slack.views.modals.base_modal.form_utils import (
        SlackFormAttributesDict,
    )

logger = logging.getLogger(__name__)


class EditMetaFormSlack(EditMetaForm):
    slack_fields: SlackFormAttributesDict = {
        "title": {
            "input": {
                "multiline": False,
                "placeholder": "Short, punchy description of what's happening.",
            },
            "block": {"hint": None},
        },
        "description": {
            "input": {
                "multiline": True,
                "placeholder": "Help people responding to the incident. This will be posted to #tech-incidents and on our internal status page.\nThis description can be edited later.",
            },
            "block": {"hint": None},
        },
        "environment": {
            "input": {
                "placeholder": "Select an environment",
            },
            "widget": {
                "label_from_instance": lambda obj: f"{obj.value} - {obj.description}",
            },
        },
    }


class EditMetaModal(ModalForm[EditMetaFormSlack]):
    open_action: str = "open_modal_incident_edit"
    update_action: str = "update_modal_incident_edit"
    push_action: str = "push_modal_incident_edit"
    open_shortcut = "modal_edit"
    callback_id: str = "incident_edit_incident"

    form_class = EditMetaFormSlack

    def build_modal_fn(self, incident: Incident, **kwargs: Any) -> View:
        blocks = self.get_form_class()(
            initial={
                "title": incident.title,
                "description": incident.description,
                "environment": incident.environment,
            },
        ).slack_blocks()

        return View(
            type="modal",
            title=f"Update incident #{incident.id}"[:24],
            submit="Update incident"[:24],
            callback_id=self.callback_id,
            private_metadata=str(incident.id),
            blocks=blocks,
        )

    def handle_modal_fn(  # type: ignore
        self, ack: Ack, body: dict[str, Any], incident: Incident, user: User
    ):

        slack_form = self.handle_form_errors(
            ack,
            body,
            forms_kwargs={
                "initial": {
                    "title": incident.title,
                    "description": incident.description,
                    "environment": incident.environment,
                }
            },
        )
        if slack_form is None:
            return
        form: EditMetaFormSlack = slack_form.form
        if len(form.cleaned_data) == 0:
            # XXX We should have a prompt for empty forms

            return
        update_kwargs: dict[str, Any] = {}
        for changed_key in form.changed_data:
            if changed_key == "environment":
                update_kwargs[f"{changed_key}_id"] = form.cleaned_data[changed_key].id
            if changed_key in {"description", "title"}:
                update_kwargs[changed_key] = form.cleaned_data[changed_key]
        if len(update_kwargs) == 0:
            logger.warning("No update to incident status")
            return
        self._trigger_incident_workflow(incident, user, **update_kwargs)

    @staticmethod
    def _trigger_incident_workflow(
        incident: Incident, user: User, **kwargs: Any
    ) -> None:
        incident.create_incident_update(created_by=user, **kwargs)


modal_edit = EditMetaModal()



from __future__ import annotations

from django.utils.timezone import datetime, get_current_timezone

from firefighter.confluence.utils import (
    CONFLUENCE_PM_ARCHIVE_TITLE_REGEX,
    CONFLUENCE_PM_TITLE_REGEX,
    parse_postmortem_title,
)


def test_postmortem_parsing_regex() -> None:
    assert (
        CONFLUENCE_PM_TITLE_REGEX.match("#20200808-inc-123 (SEV-1) This is a title")
        is not None
    )
    assert CONFLUENCE_PM_TITLE_REGEX.match("TODO: Make a postmortem") is None
    assert CONFLUENCE_PM_TITLE_REGEX.match("[ARCHIVE] Q1 2032") is None

    valid_old = "#20200808-inc-123 (SEV-1) This is a title"
    r = CONFLUENCE_PM_TITLE_REGEX.match(valid_old)
    assert r is not None
    assert r.groupdict() == {
        "date": "20200808-inc",
        "id": "123",
        "priority": "(SEV-1)",
        "title": "This is a title",
    }

    valid_recent = "#20200808-123-abc (SEV1) This is a title"
    r = CONFLUENCE_PM_TITLE_REGEX.match(valid_recent)
    assert r is not None
    assert r.groupdict() == {
        "date": "20200808",
        "id": "123",
        "priority": "(SEV1)",
        "title": "This is a title",
    }


def test_postmortem_title_parsing() -> None:
    assert parse_postmortem_title("#20200808-inc-123 (SEV-1) This is a title") == (
        datetime(2020, 8, 8, tzinfo=get_current_timezone()),
        "2020Q3",
    )


def test_postmortem_archive_title_regex() -> None:
    assert CONFLUENCE_PM_ARCHIVE_TITLE_REGEX.match("[Archive] Q1 2032") is not None
    assert CONFLUENCE_PM_ARCHIVE_TITLE_REGEX.match("Q3 2032") is not None
    assert (
        CONFLUENCE_PM_ARCHIVE_TITLE_REGEX.match("[Template] FireFighter template")
        is None
    )
    assert (
        CONFLUENCE_PM_ARCHIVE_TITLE_REGEX.match(
            "#20200808-inc-123 (SEV-1) This is a title"
        )
        is None
    )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django.dispatch.dispatcher import receiver

from firefighter.incidents.signals import postmortem_created
from firefighter.slack.messages.slack_messages import (
    SlackMessageIncidentPostMortemCreated,
)

if TYPE_CHECKING:
    from firefighter.incidents.models.incident import Incident

logger = logging.getLogger(__name__)


@receiver(signal=postmortem_created)
# pylint: disable=unused-argument
def postmortem_created_send(sender: Any, incident: Incident, **kwargs: Any) -> None:
    if not hasattr(incident, "postmortem_for"):
        logger.warning(f"No PostMortem to post for incident {incident}.")
        return

    if not hasattr(incident, "conversation"):
        logger.warning(
            f"No Incident Slack channel to post PostMortem for incident {incident}."
        )
        return
    incident.conversation.send_message_and_save(
        SlackMessageIncidentPostMortemCreated(incident), pin=True
    )
    incident.conversation.add_bookmark(
        title="Postmortem", link=incident.postmortem_for.page_url, emoji=":confluence:"
    )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Never

from django.apps import apps
from django.dispatch.dispatcher import receiver

from firefighter.confluence.models import PostMortem
from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.signals import incident_updated

if TYPE_CHECKING:
    from firefighter.incidents.models.incident import Incident
    from firefighter.incidents.models.incident_update import IncidentUpdate

if apps.is_installed("firefighter.slack"):
    from firefighter.slack.tasks.reminder_postmortem import (
        publish_fixed_next_actions,
        publish_postmortem_reminder,
    )

logger = logging.getLogger(__name__)


@receiver(signal=incident_updated)
def incident_updated_handler(
    sender: Any,
    incident: Incident,
    incident_update: IncidentUpdate,
    updated_fields: list[str],
    **kwargs: Never,
) -> None:
    if not apps.is_installed("firefighter.slack"):
        logger.error("Slack app is not installed. Skipping.")
        return
    if sender == "update_status":
        # Post postmortem reminder if needed
        if (
            "_status" in updated_fields
            and incident_update.status
            in {IncidentStatus.FIXED, IncidentStatus.POST_MORTEM}
            and incident.needs_postmortem
        ):
            if not hasattr(incident, "postmortem_for"):
                PostMortem.objects.create_postmortem_for_incident(incident)
            publish_postmortem_reminder(incident)
        elif (
            "_status" in updated_fields
            and incident_update.status
            in {IncidentStatus.FIXED, IncidentStatus.POST_MORTEM}
            and not incident.needs_postmortem
        ):
            publish_fixed_next_actions(incident)



# Generated by Django 4.2.7 on 2023-12-04 10:53

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    replaces = [
        ("jira_app", "0001_initial"),
        ("jira_app", "0002_alter_jirauser_user"),
        ("jira_app", "0003_jiraissue"),
        ("jira_app", "0004_alter_jiraissue_assignee_alter_jiraissue_reporter_and_more"),
        ("jira_app", "0005_alter_jiraissue_assignee"),
    ]

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="JiraUser",
            fields=[
                (
                    "id",
                    models.CharField(max_length=128, primary_key=True, serialize=False),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="jira_user",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Jira user",
                "verbose_name_plural": "Jira users",
            },
        ),
        migrations.CreateModel(
            name="JiraIssue",
            fields=[
                ("id", models.BigIntegerField(primary_key=True, serialize=False)),
                ("key", models.CharField(max_length=128)),
                ("issue_type", models.CharField(default="Incident", max_length=128)),
                ("project_key", models.CharField(default="INCIDENT", max_length=128)),
                ("description", models.TextField()),
                ("summary", models.CharField(help_text="Title", max_length=128)),
                (
                    "assignee",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="jira_ticket_assigned_set",
                        to="jira_app.jirauser",
                    ),
                ),
                (
                    "reporter",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="jira_ticket_reporter_set",
                        to="jira_app.jirauser",
                    ),
                ),
                (
                    "watchers",
                    models.ManyToManyField(
                        blank=True,
                        related_name="jira_ticket_watchers_set",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Jira issue",
                "verbose_name_plural": "Jira issue",
            },
        ),
    ]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django import forms
from django.conf import settings
from slack_sdk.models.blocks.blocks import SectionBlock

from firefighter.incidents.forms.create_incident import CreateIncidentForm
from firefighter.slack.slack_app import SlackApp
from firefighter.slack.views.modals.opening.set_details import SetIncidentDetails

if TYPE_CHECKING:
    from firefighter.slack.views.modals.base_modal.form_utils import (
        SlackFormAttributesDict,
    )
app = SlackApp()
logger = logging.getLogger(__name__)


class CreateIncidentFormSlack(CreateIncidentForm):
    slack_fields: SlackFormAttributesDict = {
        "title": {
            "input": {
                "multiline": False,
                "placeholder": "Short, punchy description of what's happening.",
            },
            "block": {"hint": None},
        },
        "description": {
            "input": {
                "multiline": True,
                "placeholder": "Help people responding to the incident. This will be posted to #tech-incidents and on our internal status page.\nThis description can be edited later.",
            },
            "block": {"hint": None},
        },
        "environment": {
            "input": {
                "placeholder": "Select an environment",
            },
            "widget": {
                "label_from_instance": lambda obj: f"{obj.value} - {obj.description}",
            },
        },
        "priority": {
            "input": {
                "placeholder": "Select a priority",
            },
            "widget": {
                "post_block": (
                    SectionBlock(
                        text=f"_<{settings.SLACK_SEVERITY_HELP_GUIDE_URL}|How to choose the priority?>_"
                    )
                    if settings.SLACK_SEVERITY_HELP_GUIDE_URL
                    else None
                ),
                "label_from_instance": lambda obj: f"{obj.emoji}  {obj.name} - {obj.description}",
            },
        },
        "component": {
            "input": {
                "placeholder": "Select affected issue category",
            },
        },
    }

    # Change `priority` field from parent to be a hidden field
    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.fields["priority"].widget = forms.HiddenInput()


class OpeningCriticalModal(SetIncidentDetails[CreateIncidentFormSlack]):
    open_action: str = "open_incident_critical"
    push_action: str = "push_incident_critical"
    open_shortcut = "open_incident_critical"
    callback_id: str = "open_incident_critical"

    title = "Incident Details"
    form_class = CreateIncidentFormSlack


modal_opening_critical = OpeningCriticalModal()



from __future__ import annotations

from firefighter.slack.slack_app import SlackApp
from firefighter.slack.views.events.channel_archive import channel_archive
from firefighter.slack.views.events.channel_id_changed import channel_id_changed
from firefighter.slack.views.events.channel_rename import conversation_rename
from firefighter.slack.views.events.channel_shared import channel_shared
from firefighter.slack.views.events.channel_unarchive import channel_unarchive
from firefighter.slack.views.events.channel_unshared import channel_unshared
from firefighter.slack.views.events.commands import manage_incident
from firefighter.slack.views.events.home import update_home_tab
from firefighter.slack.views.events.member_joined_channel import member_joined_channel
from firefighter.slack.views.events.member_left_channel import member_left_channel
from firefighter.slack.views.events.message import handle_message_events
from firefighter.slack.views.events.message_deleted import handle_message_deleted
from firefighter.slack.views.events.reaction_added import reaction_added

app = SlackApp()


@app.event("message")
def handle_message_events_ignore() -> None:
    """Ignore other message events.
    Must be the last event handler for message.
    """
    return



from __future__ import annotations

import sys

from decouple import Csv

from firefighter.firefighter.settings.settings_utils import config

# can't disable Slack.. yet?
ENABLE_SLACK: bool = True

SLACK_BOT_TOKEN: str = config("SLACK_BOT_TOKEN")
"""The Slack bot token to use."""
SLACK_SIGNING_SECRET: str = config("SLACK_SIGNING_SECRET")
"""The Slack signing secret to use."""

SLACK_INCIDENT_COMMAND: str = config("SLACK_INCIDENT_COMMAND", default="/incident")
"""The Slack slash command to use to create and manage incidents."""
SLACK_INCIDENT_COMMAND_ALIASES: list[str] = config(
    "SLACK_INCIDENT_COMMAND_ALIASES", cast=Csv(), default=""
)
"Comma-separated list of aliases for the incident command."

SLACK_POSTMORTEM_HELP_URL: str | None = config(
    "SLACK_POSTMORTEM_HELP_URL", default=None
)
"""URL to a guide on how to write a postmortem. Useful to point to your own documentation."""

SLACK_INCIDENT_HELP_GUIDE_URL: str | None = config(
    "SLACK_INCIDENT_HELP_GUIDE_URL", default=None
)
"URL to add in the Slack help message (/incident help). Useful to point to your own documentation."
SLACK_SEVERITY_HELP_GUIDE_URL: str | None = config(
    "SLACK_SEVERITY_HELP_GUIDE_URL", default=None
)
"URL to add in the form to choose the priority. Useful to point to your own documentation."
SLACK_EMERGENCY_COMMUNICATION_GUIDE_URL: str | None = config(
    "SLACK_EMERGENCY_COMMUNICATION_GUIDE_URL", default=None
)
"""URL to add in the Slack emergency message. Useful to point to your own documentation."""
SLACK_EMERGENCY_USERGROUP_ID: str | None = config(
    "SLACK_EMERGENCY_USERGROUP_ID", default=None
)
"""The Slack usergroup ID to use for emergency notifications. If not set, no group will be mentionned in the message."""

SLACK_APP_EMOJI: str = config("SLACK_APP_EMOJI", default=":fire_extinguisher:")
"""Emoji to represent the app in Slack surfaces. Can be an actual emoji, or a string for a custom emoji present in your Workspace, like ":incident_logo:"."""

_is_default_skip_check_cmd = "generate_manifest" in sys.argv
FF_SLACK_SKIP_CHECKS: bool = config(
    "FF_SLACK_SKIP_CHECKS", cast=bool, default=_is_default_skip_check_cmd
)
"""Skip Slack checks. Only use for testing or demo."""



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from celery import shared_task
from django.conf import settings

from firefighter.slack.slack_templating import user_slack_handle_or_name

if TYPE_CHECKING:
    from celery import Signature

    from firefighter.incidents.models.user import User

if settings.ENABLE_PAGERDUTY:
    from firefighter.pagerduty.models import PagerDutyOncall
    from firefighter.pagerduty.tasks import fetch_oncalls
if settings.ENABLE_CONFLUENCE:
    from firefighter.confluence.service import confluence_service
if settings.ENABLE_SLACK:
    from firefighter.slack.models.conversation import Conversation
    from firefighter.slack.models.user import SlackUser

logger = logging.getLogger(__name__)

BASE_URL: str = settings.BASE_URL


@shared_task(name="incidents.update_oncall")
def update_oncall() -> None:
    """Fetch current on-calls and update the on-call Slack topic and Confluence page."""
    chain: Signature[bool] = (
        fetch_oncalls.s()  # pyright: ignore[reportUnboundVariable]
        | update_oncall_views.s()
    )
    chain()


@shared_task(name="incidents.update_oncall_views")
def update_oncall_views(*_args: Any, **_kwargs: Any) -> bool:
    """Updates the on-call Slack topic and Confluence page containing the info for the on-call personnel."""
    if not settings.ENABLE_PAGERDUTY:
        logger.error("Can't update on-call users without PagerDuty enabled.")
        return False

    oncall_users_grouped_per_ep = PagerDutyOncall.objects.get_current_oncalls_per_escalation_policy_name_first_responder()

    # Check that we have Slack ID and handle for these users, if needed
    if settings.ENABLE_SLACK:
        for user in oncall_users_grouped_per_ep.values():
            if not hasattr(user, "slack_user"):
                SlackUser.objects.add_slack_id_to_user(user)
                logger.debug("Added Slack ID to user: %s", user)
        update_oncall_slack_topic(oncall_users_grouped_per_ep)
    else:
        logger.warning("Not updating on-call Slack topic, Slack integration disabled.")

    update_oncall_confluence(oncall_users_grouped_per_ep)
    logger.info("Updated on-call users on Confluence and Slack topic.")
    return True


def create_oncall_topic(users: dict[str, User]) -> str:
    base_topic = f":uk: <{BASE_URL}|IT PRD SEV[1-3] incidents> - On-call: "
    for service_name, user in users.items():
        user_text = f"{service_name.upper()} "
        user_text += user_slack_handle_or_name(user) + " "
        if (
            hasattr(user, "pagerduty_user")
            and user.pagerduty_user
            and user.pagerduty_user.phone_number != ""
        ):
            user_text += f"<tel:+{user.pagerduty_user.phone_number}> "
        base_topic += user_text
    if len(base_topic) > 250:
        logger.warning(
            f'Tech channel on-call Slack topic is too long and will be cut at 250 chars: "{base_topic}"'
        )
    logger.debug(f"New on-call topic: {base_topic}.")
    return base_topic[:250]


def update_oncall_slack_topic(
    users: dict[str, User],
) -> bool:
    """TODO(gab) Move in the Slack app."""
    if not settings.ENABLE_SLACK:
        return False

    oncall_topic = create_oncall_topic(users)
    tech_incidents_conversation = Conversation.objects.get_or_none(tag="tech_incidents")
    if tech_incidents_conversation:
        res = tech_incidents_conversation.update_topic(oncall_topic)
        if res.get("ok"):
            logger.info(
                f"Updated on-call Slack topic on {tech_incidents_conversation}."
            )
            return True
    else:
        logger.warning(
            "Could not find tech_incidents conversation! Is there a channel with tag tech_incidents?"
        )
        return False
    logger.error("Failed to update on-call Slack topic.")
    return False


def update_oncall_confluence(users: dict[str, User]) -> bool:
    if not settings.ENABLE_CONFLUENCE:
        return False

    return confluence_service.update_oncall_page(users)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.db import models
from django_stubs_ext.db.models import TypedModelMeta

from firefighter.incidents.models.user import User

if TYPE_CHECKING:
    from firefighter.incidents.models.incident import Incident  # noqa: F401
    from firefighter.incidents.models.incident_role_type import (  # noqa: F401
        IncidentRoleType,
    )


class IncidentMembership(models.Model):
    incident = models.ForeignKey(
        "Incident", on_delete=models.CASCADE, db_index=True
    )
    user = models.ForeignKey(User, on_delete=models.CASCADE, db_index=True)

    class Meta(TypedModelMeta):
        constraints = [
            models.UniqueConstraint(
                fields=["incident", "user"], name="unique_incident_membership"
            ),
        ]

    def __str__(self) -> str:
        return f"{self.user} - {self.incident}"


class IncidentRole(models.Model):
    incident = models.ForeignKey(
        "Incident", on_delete=models.CASCADE, related_name="roles_set", db_index=True
    )
    user = models.ForeignKey(
        "User", on_delete=models.CASCADE, related_name="roles_set", db_index=True
    )
    role_type = models.ForeignKey(
        "IncidentRoleType",
        on_delete=models.CASCADE,
        related_name="roles_set",
        db_index=True,
    )
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["incident", "user", "role_type"], name="unique_incident_role"
            ),
        ]

    def __str__(self) -> str:
        return f"{self.user} - {self.incident} - {self.role_type}"



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from celery import shared_task
from django.conf import settings
from pdpyras import PDClientError, PDHTTPError

from firefighter.incidents.models.incident import Incident
from firefighter.pagerduty.models import PagerDutyIncident, PagerDutyService
from firefighter.pagerduty.service import pagerduty_service

if TYPE_CHECKING:
    from firefighter.incidents.models.user import User
logger = logging.getLogger(__name__)
APP_DISPLAY_NAME: str = settings.APP_DISPLAY_NAME


@shared_task(name="pagerduty.trigger_oncall")
def trigger_oncall(
    oncall_service: PagerDutyService,
    title: str,
    details: str,
    incident_key: str,
    conference_url: str,
    incident_id: int | None = None,
    triggered_by: User | None = None,
) -> PagerDutyIncident:
    """Celery task to trigger an on-call in PagerDuty, from a FireFighter incident.

    XXX Trigger from PD user if it exists, instead of admin.
    XXX Should be a service ID instead of a service object.
    """
    service = oncall_service
    if incident_id:
        incident = Incident.objects.get(id=incident_id)
        details = f"""Triggered from {APP_DISPLAY_NAME} incident #{incident.id} {f"by {triggered_by.full_name}" if triggered_by else ""}
Priority: {incident.priority}
Environment: {incident.environment}
Component: {incident.component}
FireFighter page: {incident.status_page_url + "?utm_medium=FireFighter+PagerDuty&utm_source=PagerDuty+Incident&utm_campaign=OnCall+Message+In+Channel"}
Slack channel #{incident.slack_channel_name}: {incident.slack_channel_url}

Incident Details:
{details}
"""
    try:
        res = pagerduty_service.client.create_incident(
            title=title,
            pagerduty_id=service.pagerduty_id,
            details=details,
            incident_key=incident_key,
            conference_url=conference_url,
        )

    except PDHTTPError as e:
        if e.response.status_code == 404:
            logger.exception("User not found")
        else:
            logger.exception("Transient network error: %s", e.msg)
            raise
    except PDClientError as e:
        logger.exception("Non-transient network or client error: %s", e.msg)
    # TODO Error handling

    if not 200 <= res.status_code < 300:
        logger.error({
            "message": "Error when calling PagerDuty API",
            "request": res.request.__dict__.get("body"),
            "response": res.json(),
        })
        err_msg = f"Error when calling PagerDuty API. {res.json()}"
        raise ValueError(err_msg)

    pd_incident = res.json()["incident"]

    pd_incident_db, _ = PagerDutyIncident.objects.update_or_create(
        incident_key=pd_incident["incident_key"],
        defaults={
            "title": pd_incident["title"][:128],
            "urgency": pd_incident["urgency"][:128],
            "incident_number": pd_incident["incident_number"],
            "status": pd_incident["status"][:128],
            "service_id": service.id,
            "details": pd_incident["body"]["details"][:3000],  # get_in
            "summary": pd_incident["summary"][:256],  # get_in
            "web_url": pd_incident["html_url"][:256],
            "api_url": pd_incident["self"][:256],
            "incident_id": incident_id,
        },
    )

    return pd_incident_db



from __future__ import annotations

from datetime import UTC, datetime, timedelta
from typing import Any

from django.utils.translation import ngettext as _
from slack_sdk.models.blocks.basic_components import MarkdownTextObject
from slack_sdk.models.blocks.blocks import (
    Block,
    SectionBlock,
)
from slack_sdk.models.views import View

from firefighter.incidents.models.incident import Incident
from firefighter.slack.views.modals.base_modal.base import SlackModal


class CheckCurrentIncidentsModal(
    SlackModal,
):
    handle_modal_fn = None
    open_action = "open_check_current_incidents"
    push_action: str = "push_check_current_incidents"
    callback_id = "check_current_incidents"

    @staticmethod
    def build_modal_fn(body: dict[str, Any], **kwargs: Any) -> View:
        # ruff: noqa: PLC0415
        from firefighter.slack.views.events.home import _home_incident_element

        # Close or back button?
        view: dict[str, Any] = body.get("view", {})
        close_text = (
            "Back"
            if view.get("root_view_id") or not view.get("clear_on_close")
            else "Close"
        )

        incidents = list(
            Incident.objects.filter(

                created_at__gte=datetime.now(UTC) - timedelta(hours=1)
            )
            .order_by("-created_at")
            .select_related(
                "priority",
                "component",
                "environment",
                "component__group",
                "conversation",
            )[:30]
        )
        if len(incidents) == 0:
            blocks: list[Block] = [
                SectionBlock(
                    text=MarkdownTextObject(text="No incidents in the past hour.")
                )
            ]
        else:
            blocks = [
                SectionBlock(
                    text=MarkdownTextObject(
                        text=_(
                            "Below is the incident that has been created in the past hour.",
                            "Below are the incidents that have been created in the past hour.",
                            len(incidents),
                        )
                    )
                )
            ]
        for incident in incidents:
            blocks.extend(_home_incident_element(incident, show_actions=False))

        return View(
            type="modal",
            title="Last critical incidents"[:24],
            blocks=blocks,
            close=close_text,
        )


modal_checker = CheckCurrentIncidentsModal()



from __future__ import annotations

import logging
from unittest.mock import MagicMock

import pytest

from firefighter.incidents.factories import IncidentFactory, UserFactory
from firefighter.incidents.models import Incident
from firefighter.slack.views import SendSosModal

logger = logging.getLogger(__name__)


@pytest.mark.django_db
class TestSendSosModal:
    @staticmethod
    @pytest.fixture
    def incident() -> Incident:
        """Returns a valid incident."""
        return IncidentFactory.build()

    @staticmethod
    def test_send_sos_modal_build(incident: Incident) -> None:
        modal = SendSosModal()
        res = modal.build_modal_fn(incident)

        # To dict will validate the JSON as well (e.g. check the view title length)
        # Any validation error will raise an exception
        assert res.to_dict()

        values = res.to_dict()
        assert "blocks" in values
        assert len(values["blocks"]) == 4

    @staticmethod
    def test_submit_empty_bodied_form(incident: Incident) -> None:
        modal = SendSosModal()
        ack = MagicMock()
        user = UserFactory.build()
        with pytest.raises(TypeError, match="Expected a values dict in the body"):
            modal.handle_modal_fn(ack=ack, body={}, incident=incident, user=user)



from __future__ import annotations

from django.contrib import admin

import firefighter.incidents.admin
from firefighter.confluence.models import ConfluencePage, PostMortem, Runbook
from firefighter.incidents.models.incident import Incident


@admin.register(ConfluencePage)
class ConfluencePageAdmin(admin.ModelAdmin[ConfluencePage]):
    model = ConfluencePage

    list_display: tuple[str, ...] = (
        "name",
        "page_id",
        "created_at",
        "updated_at",
    )
    search_fields: tuple[str, ...] = (
        "name",
        "page_id",
    )
    ordering: tuple[str, ...] = ("-created_at",)
    list_display_links = ["name", "page_id"]


class PostMortemAdminInline(admin.StackedInline[PostMortem, Incident]):
    model = PostMortem
    extra = 0
    verbose_name = "Confluence Post-Mortem"


@admin.register(PostMortem)
class PostMortemAdmin(ConfluencePageAdmin):
    model = PostMortem
    search_fields = ("name", "page_id", "incident__id")
    list_display = ("incident_id", "name", "page_id")
    ordering = ("-incident_id",)


@admin.register(Runbook)
class RunbookAdmin(ConfluencePageAdmin):
    model = Runbook
    list_display = ("service_type", "service_name", "title", "page_id")
    ordering = ("service_type", "service_name")
    list_display_links = ["service_type", "service_name", "title", "page_id"]
    list_filter = ("service_type",)


# Add inlines to incidents models
firefighter.incidents.admin.incident_inlines.append(PostMortemAdminInline)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django.conf import settings
from rest_framework import serializers

from firefighter.incidents.models.user import User
from firefighter.jira_app.client import (
    JiraAPIError,
    JiraUserNotFoundError,
    SlackNotificationError,
)
from firefighter.raid.client import client as jira_client
from firefighter.raid.forms import (
    alert_slack_comment_ticket,
    alert_slack_new_jira_ticket,
    alert_slack_update_ticket,
)
from firefighter.raid.models import JiraTicket
from firefighter.raid.utils import get_domain_from_email
from firefighter.slack.models.user import SlackUser

if TYPE_CHECKING:
    from firefighter.jira_app.models import JiraUser


JIRA_USER_IDS: dict[str, str] = settings.RAID_JIRA_USER_IDS

logger = logging.getLogger(__name__)


class IgnoreEmptyStringListField(serializers.ListField):
    def to_internal_value(self, data: list[Any] | Any) -> list[str]:
        # Check if data is a list
        if not isinstance(data, list):
            msg = f'Expected a list but got type "{type(data).__name__}".'
            raise serializers.ValidationError(msg)

        # Filter out any empty string values
        data = [item for item in data if item != ""]

        return super().to_internal_value(data)


def validate_no_spaces(value: str) -> None:
    """Ensure the string does not contain spaces."""
    if " " in value:
        raise serializers.ValidationError("The string cannot contain spaces.")


def get_reporter_user_from_email(reporter_email: str) -> tuple[User, JiraUser, str]:
    user_domain: str = get_domain_from_email(reporter_email)
    try:
        reporter_user: User = User.objects.get(email=reporter_email)
        reporter = jira_client.get_jira_user_from_user(reporter_user)
    except (User.DoesNotExist, JiraUserNotFoundError):
        reporter_user_tmp = SlackUser.objects.upsert_by_email(email=reporter_email)
        if reporter_user_tmp is not None:
            reporter_user = reporter_user_tmp

        logger.warning(
            f"Reporter email {reporter_email} not found in Jira. Using default user."
        )

        # Get default Jira user, depending on the domain
        # We might have accounts with Slack users, but not Jira users
        # In that case, we use the default Jira user for that domain
        match JIRA_USER_IDS.get(user_domain):
            case jira_user_id if jira_user_id is not None:
                reporter = jira_client.get_jira_user_from_jira_id(jira_user_id)
                reporter_user = (
                    reporter.user if reporter_user_tmp is None else reporter_user_tmp
                )
            case _:
                reporter = jira_client.get_jira_user_from_jira_id(
                    settings.RAID_DEFAULT_JIRA_QRAFT_USER_ID
                )
                reporter_user = (
                    reporter.user if reporter_user_tmp is None else reporter_user_tmp
                )
    return reporter_user, reporter, user_domain


class LandbotIssueRequestSerializer(serializers.ModelSerializer[JiraTicket]):
    id = serializers.IntegerField(read_only=True)
    key = serializers.CharField(read_only=True)
    seller_contract_id = serializers.CharField(
        max_length=128,
        write_only=True,
        allow_null=True,
        allow_blank=True,
    )
    zoho = serializers.CharField(
        max_length=256,
        write_only=True,
        allow_null=True,
        allow_blank=True,
    )
    platform = serializers.ChoiceField(
        write_only=True, choices=["ES", "IT", "FR", "UK", "DE", "All", "Internal"]
    )
    reporter_email = serializers.EmailField(write_only=True)

    labels = serializers.ListField(
        required=False,
        write_only=True,
        child=serializers.CharField(
            max_length=128,
            allow_blank=False,
            allow_null=False,
            validators=[validate_no_spaces],
        ),
        help_text="List of labels to be added to the ticket. Labels cannot contain spaces and must not exceed 255 characters.",
    )
    impacted_area = serializers.CharField(
        max_length=128,
        write_only=True,
        allow_null=True,
        allow_blank=True,
    )
    suggested_team_routing = serializers.CharField(max_length=10, write_only=True)
    project = serializers.ChoiceField(
        required=False,
        allow_blank=True,
        allow_null=True,
        default="SBI",
        choices=[
            "SBI",
        ],
    )
    priority = serializers.IntegerField(
        min_value=1, max_value=4, write_only=True, allow_null=True
    )
    business_impact = serializers.ChoiceField(
        write_only=True, choices=["High", "Medium", "Low"], allow_null=True
    )
    environments = serializers.ListField(
        write_only=True,
        allow_null=True,
        default=["-"],
        child=serializers.ChoiceField(
            allow_null=True,
            allow_blank=True,
            choices=["INT", "STG", "PRD", "SUPPORT", "-"],
        ),
    )
    attachments = serializers.CharField(
        required=False,
        allow_blank=True,
        allow_null=True,
    )
    issue_type = serializers.ChoiceField(
        required=False,
        allow_blank=True,
        allow_null=True,
        default="Incident",
        choices=[
            "Incident",
            "Service Request",
            "Documentation/Process Request",
            "Feature Request",
        ],
    )

    def validate_environments(self, value: list[str] | None) -> list[str] | Any:
        if not value:
            return self.fields["environments"].default
        return value

    def create(self, validated_data: dict[str, Any]) -> JiraTicket:
        reporter_email: str = validated_data["reporter_email"]

        reporter_user, reporter, user_domain = get_reporter_user_from_email(
            reporter_email
        )
        user_in_description: str = (
            f"\nReporter email {reporter_email}\n"
            if "manomano" not in user_domain
            else ""
        )

        issue = jira_client.create_issue(
            issuetype=validated_data["issue_type"],
            summary=validated_data["summary"],
            description=f"""{validated_data.get("description")}\n
\n
🤖 This ticket was created through the Incident Verification Bot https://landbot.pro/v3/H-1332620-8BF395J9STSHL7IN/index.html.\n
\n
{user_in_description}
\n
""",
            assignee=None,
            reporter=reporter.id,
            labels=validated_data["labels"],
            priority=validated_data["priority"],
            seller_contract_id=validated_data["seller_contract_id"],
            zoho_desk_ticket_id=validated_data["zoho"],
            platform=validated_data["platform"],
            area=validated_data["impacted_area"],
            business_impact=validated_data["business_impact"],
            environments=validated_data["environments"],
            suggested_team_routing=validated_data["suggested_team_routing"],
            project=validated_data["project"],
        )
        issue_id = issue.get("id")
        if issue_id is None:
            logger.error("Could not create Jira ticket")
            raise JiraAPIError("Could not create Jira ticket")
        if validated_data["attachments"] is not None:
            #  Prepare attachments and double check we don't have any empty strings
            attachments: list[str] = [
                a
                for a in validated_data["attachments"]
                .replace("[", "")
                .replace("]", "")
                .replace("'", "")
                .split(", ")
                if a
            ]
            jira_client.add_attachments_to_issue(issue_id, attachments)

        jira_ticket = JiraTicket.objects.create(**issue)

        # Send messages in the relevant Slack channels if needed and alert the reporter in DMs
        alert_slack_new_jira_ticket(
            jira_ticket, reporter_user=reporter_user, reporter_email=reporter_email
        )

        return jira_ticket

    class Meta:
        model = JiraTicket
        fields = [
            "summary",
            "description",
            "seller_contract_id",
            "zoho",
            "platform",
            "reporter_email",
            "impacted_area",
            "labels",
            "environments",
            "issue_type",
            "business_impact",
            "priority",
            "id",
            "key",
            "attachments",
            "suggested_team_routing",
            "project",
        ]


class JiraWebhookUpdateSerializer(serializers.Serializer[Any]):
    issue = serializers.DictField()
    changelog = serializers.DictField()
    user = serializers.DictField()
    webhookEvent = serializers.ChoiceField(  # noqa: N815
        write_only=True,
        choices=["jira:issue_updated"],
    )

    def create(self, validated_data: dict[str, Any]) -> bool:
        jira_field_modified = validated_data["changelog"].get("items")[0].get("field")
        if jira_field_modified in {
            "Priority",
            "project",
            "description",
            "status",
        }:
            jira_ticket_key = validated_data["issue"].get("key")
            status = alert_slack_update_ticket(
                jira_ticket_id=validated_data["issue"].get("id"),
                jira_ticket_key=jira_ticket_key,
                jira_author_name=validated_data["user"].get("displayName"),
                jira_field_modified=jira_field_modified,
                jira_field_from=validated_data["changelog"]
                .get("items")[0]
                .get("fromString"),
                jira_field_to=validated_data["changelog"]
                .get("items")[0]
                .get("toString"),
            )
            if status is not True:
                logger.error(
                    f"Could not alert in Slack for the update/s in the Jira ticket {jira_ticket_key}"
                )
                raise SlackNotificationError("Could not alert in Slack")

        return True

    def update(self, instance: Any, validated_data: Any) -> Any:
        raise NotImplementedError


class JiraWebhookCommentSerializer(serializers.Serializer[Any]):
    issue = serializers.DictField()
    comment = serializers.DictField()
    webhookEvent = serializers.ChoiceField(  # noqa: N815
        write_only=True,
        choices=["comment_created", "comment_updated", "comment_deleted"],
    )

    def create(self, validated_data: dict[str, Any]) -> bool:
        jira_ticket_key = validated_data["issue"].get("key")
        status = alert_slack_comment_ticket(
            webhook_event=validated_data["webhookEvent"],
            jira_ticket_id=validated_data["issue"].get("id"),
            jira_ticket_key=jira_ticket_key,
            author_jira_name=validated_data["comment"].get("author").get("displayName"),
            comment=validated_data["comment"].get("body"),
        )
        if status is not True:
            logger.error(
                f"Could not alert in Slack for created/modified comment in the Jira ticket {jira_ticket_key}"
            )
            raise SlackNotificationError("Could not alert in Slack")
        return True

    def update(self, instance: Any, validated_data: Any) -> Any:
        raise NotImplementedError



from __future__ import annotations

import json
import logging
from datetime import timedelta
from typing import TYPE_CHECKING, Any, Literal, cast

from django.conf import settings
from django.utils import timezone
from django.utils.translation import gettext as _
from django.utils.translation import ngettext
from slack_sdk.models.blocks.basic_components import MarkdownTextObject, Option
from slack_sdk.models.blocks.block_elements import ButtonElement, StaticSelectElement
from slack_sdk.models.blocks.blocks import (
    ActionsBlock,
    Block,
    ContextBlock,
    DividerBlock,
    SectionBlock,
)
from slack_sdk.models.views import View

from firefighter.firefighter.utils import is_during_office_hours
from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.forms.select_impact import SelectImpactForm
from firefighter.incidents.models.impact import ImpactType
from firefighter.incidents.models.incident import Incident
from firefighter.incidents.models.priority import Priority
from firefighter.slack.slack_app import SlackApp
from firefighter.slack.slack_incident_context import get_user_from_context
from firefighter.slack.views.modals.base_modal.base import SlackModal
from firefighter.slack.views.modals.base_modal.form_utils import SlackFormJSONEncoder
from firefighter.slack.views.modals.base_modal.modal_utils import update_modal
from firefighter.slack.views.modals.opening.check_current_incidents import (
    CheckCurrentIncidentsModal,
)
from firefighter.slack.views.modals.opening.details.critical import OpeningCriticalModal
from firefighter.slack.views.modals.opening.types import OpeningData, ResponseType

if TYPE_CHECKING:
    from slack_bolt.context.ack.ack import Ack

    from firefighter.incidents.forms.create_incident import CreateIncidentFormBase
    from firefighter.incidents.models.user import User
    from firefighter.slack.views.modals.opening.set_details import SetIncidentDetails

app = SlackApp()
logger = logging.getLogger(__name__)

SLACK_SEVERITY_HELP_GUIDE_URL: str | None = settings.SLACK_SEVERITY_HELP_GUIDE_URL


INCIDENT_TYPES: dict[ResponseType, dict[str, dict[str, Any]]] = {
    "critical": {
        "critical": {
            "label": "Critical",
            "slack_form": OpeningCriticalModal,
        },
    },
}


class OpenModal(SlackModal):
    open_action: str = "open_incident"
    open_shortcut = "open_incident"
    callback_id: str = "incident_open"

    def build_modal_fn(
        self, open_incident_context: OpeningData | None = None, user: User | None = None
    ) -> View:
        if user is None:
            raise ValueError("user is required for OpenModal!")
        open_incident_context = open_incident_context or OpeningData()
        # 1. Check if impact form is good
        is_impact_form_valid: bool = self._check_impact_form(open_incident_context)

        # 2. Check if we have a normal incident type
        incident_type_value: str | None = open_incident_context.get(
            "incident_type", None
        )

        # 3. Check if details form is good
        (
            is_details_form_valid,
            details_form_class,
            details_form_modal_class,
            details_form,
        ) = self._check_details_form(open_incident_context, incident_type_value)

        # Add priority field to details form from priority
        # XXX Priority should be on both impact_form_data and details_form_data and not in OpeningData
        if (
            open_incident_context.get("priority")
            and "priority" in open_incident_context
            and open_incident_context["priority"]
        ):
            if (
                "details_form_data" not in open_incident_context
                or open_incident_context["details_form_data"] is None
            ):
                open_incident_context["details_form_data"] = {}
            assert open_incident_context["details_form_data"] is not None  # noqa: S101
            open_incident_context["details_form_data"]["priority"] = (
                open_incident_context["priority"]
            )

        # Make blocks with all this context
        intro_blocks = self.get_intro_blocks()
        set_impacts_blocks = self.get_set_impact_blocks(
            open_incident_context, impact_form_done=is_impact_form_valid
        )
        choose_response_type_blocks = self._build_response_type_blocks(
            open_incident_context
        )
        select_incident_type_blocks = self.get_select_incident_type_blocks(
            open_incident_context, incident_type_value
        )
        set_details_blocks = self.get_set_details_blocks(
            open_incident_context,
            details_form_done=is_details_form_valid,
            details_form_modal_class=details_form_modal_class,
        )

        # Can we submit?
        can_submit: bool = is_impact_form_valid and is_details_form_valid
        done_review_blocks = self.get_done_review_blocks(
            open_incident_context,
            user,
            is_details_form_valid,
            details_form_class,
            details_form,
            can_submit,
        )
        blocks: list[Block] = [
            *intro_blocks,
            DividerBlock(),
            *set_impacts_blocks,
            *choose_response_type_blocks,
            *select_incident_type_blocks,
            *set_details_blocks,
            *done_review_blocks,
        ]
        return View(
            type="modal",
            title="Create an incident"[:24],
            submit="Create the incident"[:24] if can_submit else None,
            callback_id=self.callback_id,
            blocks=blocks,
            clear_on_close=False,
            close=None,
        )

    @staticmethod
    def get_intro_blocks() -> list[Block]:
        blocks: list[Block] = [
            SectionBlock(
            text=(
                "Hello and thanks for reporting a new incident! :beetle:\n\n"
                "Please report as much information as you can!\n\n"
                "More information about the incident process in this "
                "<https://manomano.atlassian.net/wiki/spaces/TC/pages/3928261283/IMPACT+-+Incident+Management+Platform|documentation>."
            )
            )
        ]

        recent_critical_incidents: int = Incident.objects.filter(
            created_at__gte=timezone.now() - timedelta(hours=1)
        ).count()
        if recent_critical_incidents > 0:
            blocks.append(
                OpenModal.warning_recent_critical_incidents(recent_critical_incidents)
            )

        return blocks

    @staticmethod
    def get_set_impact_blocks(
        open_incident_context: OpeningData, *, impact_form_done: bool
    ) -> list[Block]:
        from firefighter.slack.views.modals.opening.select_impact import (  # noqa: PLC0415
            SelectImpactModal,
        )

        return [
            SectionBlock(
                text=f"{'✅' if impact_form_done else '📝'} First, define the incident impacts and priority.",
                accessory=ButtonElement(
                    text="Edit impacts" if impact_form_done else "Set impacts",
                    action_id=SelectImpactModal.push_action,
                    value=json.dumps(open_incident_context, cls=SlackFormJSONEncoder),
                ),
            ),
        ]

    @staticmethod
    def warning_recent_critical_incidents(count: int) -> SectionBlock:
        return SectionBlock(
            text=ngettext(
                "> 🚨 A critical incident has been created in the past hour. Please check it to avoid duplicate incidents.",
                "> 🚨 %(count)d critical incidents have been created in the past hour. Please check them to avoid duplicate incidents.",
                count,
            )
            % {"count": count},
            accessory=ButtonElement(
                text="Check incidents",
                action_id=CheckCurrentIncidentsModal.push_action,
            ),
        )

    @staticmethod
    def get_set_details_blocks(
        open_incident_context: OpeningData,
        *,
        details_form_done: bool,
        details_form_modal_class: type[SetIncidentDetails[Any]] | None,
    ) -> list[Block]:
        if details_form_modal_class is None:
            return []
        return [
            SectionBlock(
                text=f"{'✅' if details_form_done else '📝'} Finally, add incident details",
                accessory=ButtonElement(
                    text="Edit details" if details_form_done else "Set details",
                    action_id=details_form_modal_class.push_action,
                    value=json.dumps(open_incident_context, cls=SlackFormJSONEncoder),
                ),
            ),
        ]

    @staticmethod
    def get_select_incident_type_blocks(
        open_incident_context: OpeningData,
        incident_type_value: str | None,
    ) -> list[Block]:
        response_type = open_incident_context.get("response_type")
        if (
            response_type is None
            or response_type not in INCIDENT_TYPES
            or len(INCIDENT_TYPES[response_type]) == 1
        ):
            return []
        options = [
            Option(value=incident_type, label=incident_type_data["label"])
            for incident_type, incident_type_data in INCIDENT_TYPES[
                response_type
            ].items()
        ]
        initial_option = next(
            filter(lambda x: x.value == incident_type_value, options), None
        )
        return [
            SectionBlock(
                text=("✅" if incident_type_value else "📝")
                + _("Then, select the type of issue / affected users"),
                accessory=StaticSelectElement(
                    placeholder=_("Set type"),
                    action_id="set_type",
                    options=options,
                    initial_option=initial_option,
                ),
            ),
        ]

    @staticmethod
    def get_done_review_blocks(
        open_incident_context: OpeningData,
        user: User,
        details_form_done: bool,  # noqa: FBT001
        details_form_class: type[CreateIncidentFormBase] | None,
        details_form: CreateIncidentFormBase | None,
        can_submit: bool,  # noqa: FBT001
    ) -> list[Block]:
        if not can_submit:
            return []
        done_review_blocks: list[Block] = [
            DividerBlock(),
            SectionBlock(
                text=":tada: Almost done! Click the button below to create the incident."
            ),
        ]

        if open_incident_context.get("response_type") == "critical":
            slack_msg = None
            if details_form_done and details_form_class and details_form:
                incident = Incident(
                    status=IncidentStatus.OPEN,  # type: ignore
                    created_by=user,
                    **details_form.cleaned_data,
                )
                users_list: set[User] = {*incident.build_invite_list(), user}
                slack_msg = f"> :slack: A dedicated Slack channel will be created, and around {len(users_list)} responders will be invited to help.\n"

            if slack_msg is None:
                slack_msg = "> :slack: A dedicated Slack channel will be created, and responders will be invited to help.\n"
            text = f"> :firefighter_incident: This will trigger a critical incident response.\n{slack_msg}> :jira_new: An associated Jira ticket will also be created."
            if not is_during_office_hours(timezone.now()):
                text += "\n> :pagerduty: If you need it, you'll be able to escalate the incident to our 24/7 on-call response teams."
        else:
            text = "> :raid_logo: This will trigger a normal incident response.\n> :jira_new: A Jira ticket will be created."
        done_review_blocks += [SectionBlock(text=text)]

        return done_review_blocks

    @staticmethod
    def _check_impact_form(open_incident_context: OpeningData) -> bool:
        done = bool(
            open_incident_context and open_incident_context.get("impact_form_data")
        )
        if not done:
            return False
        form = SelectImpactForm(open_incident_context.get("impact_form_data"))
        return form.is_valid()

    def _check_details_form(
        self,
        open_incident_context: OpeningData,
        incident_type_value: str | None,
    ) -> tuple[
        bool,
        type[CreateIncidentFormBase] | None,
        type[SetIncidentDetails[Any]] | None,
        CreateIncidentFormBase | None,
    ]:
        details_form_done = False
        details_form_class = None
        details_form = None

        details_form_modal_class = self.get_details_modal_form_class(
            open_incident_context, incident_type_value
        )

        if (
            open_incident_context.get("details_form_data")
            and "details_form_data" in open_incident_context
            and open_incident_context["details_form_data"] is not None
        ):
            (
                details_form_done,
                details_form_class,
                details_form,
            ) = self._validate_details_form(
                details_form_modal_class, open_incident_context["details_form_data"]
            )

        return (
            details_form_done,
            details_form_class,
            details_form_modal_class,
            details_form,
        )

    @staticmethod
    def _validate_details_form(
        details_form_modal_class: type[SetIncidentDetails[Any]] | None,
        details_form_data: dict[str, Any],
    ) -> tuple[
        bool, type[CreateIncidentFormBase] | None, CreateIncidentFormBase | None
    ]:
        if not details_form_modal_class:
            return False, None, None

        details_form_class: type[CreateIncidentFormBase] = (
            details_form_modal_class.form_class
        )

        if not details_form_class:
            return False, None, None

        details_form: CreateIncidentFormBase = details_form_class(details_form_data)
        is_valid = details_form.is_valid()

        return is_valid, details_form_class, details_form

    @staticmethod
    def _build_response_type_blocks(open_incident_context: OpeningData) -> list[Block]:
        selected_response_type = open_incident_context.get("response_type")
        if selected_response_type not in {"critical", "normal"}:
            return []

        response_types: list[ResponseType] = cast(
            "list[ResponseType]", INCIDENT_TYPES.keys()
        )
        elements: list[ButtonElement] = []

        for response_type in response_types:
            if response_type != selected_response_type:
                continue

            is_selected = (
                open_incident_context.get("response_type") == response_type
                or len(INCIDENT_TYPES) == 1
            )
            style: str | None = "primary" if is_selected else None
            text = (
                ":slack: Slack :jira_new: Jira ticket"
                if response_type == "critical"
                else ":jira_new: Jira ticket"
            )
            button = ButtonElement(
                text=text,
                action_id=f"incident_open_set_res_type_{response_type}",
                value=json.dumps(open_incident_context, cls=SlackFormJSONEncoder),
                style=style,
            )
            elements.append(button)

        blocks: list[Block] = [ActionsBlock(elements=elements)]
        if impact_form_data := open_incident_context.get("impact_form_data"):
            impact_form = SelectImpactForm(impact_form_data)
            if impact_form.is_valid():
                priority: Priority = Priority.objects.get(
                    value=impact_form.suggest_priority_from_impact()
                )
                process = ":slack: Slack :jira_new: Jira ticket" if open_incident_context.get("response_type") == "critical" else ":jira_new: Jira ticket"

                impact_descriptions = OpenModal._get_impact_descriptions(open_incident_context)
                blocks.append(
                    ContextBlock(
                        elements=[
                            MarkdownTextObject(
                                text=f"> {priority.emoji} Selected priority: *{priority} - {priority.description}*\n"
                                f"> ⏱️ SLA: {priority.sla}\n"
                                f"> :gear: Process: {process}\n"
                                f"> :pushpin: Selected impacts:\n"
                                f"{impact_descriptions}\n"
                                + (
                                    (
                                        "\n> Critical incidents are for *emergency* only"
                                        + (
                                            f" <{SLACK_SEVERITY_HELP_GUIDE_URL}|learn more>"
                                            if SLACK_SEVERITY_HELP_GUIDE_URL
                                            else "."
                                        )
                                    )
                                    if selected_response_type == "critical"
                                    else ""
                                )
                            )
                        ]
                    ),
                )

                if (
                    priority.recommended_response_type
                    and priority.recommended_response_type != selected_response_type
                ):
                    blocks.append(
                        ContextBlock(
                            elements=[
                                MarkdownTextObject(
                                    text=f"> :warning: The current priority ({priority}) is usually handled with a {priority.recommended_response_type} response type."
                                )
                            ]
                        )
                    )

        return blocks

    @staticmethod
    def _get_impact_descriptions(open_incident_context: OpeningData) -> str:
        impact_form_data = open_incident_context.get("impact_form_data", {})
        impact_descriptions = ""
        if impact_form_data:
            for value in impact_form_data.values():
                if value.name != "NO" and value.description:
                    if hasattr(value, "impact_type_id") and value.impact_type_id:
                        impact_type = ImpactType.objects.get(pk=value.impact_type_id)
                        if impact_type:
                            impact_descriptions += f"> \u00A0\u00A0 :exclamation: {impact_type} - {value}\n"
                    for line in str(value.description).splitlines():
                        impact_descriptions += f"> \u00A0\u00A0\u00A0\u00A0\u00A0\u00A0 • {line}\n"
        return impact_descriptions

    @staticmethod
    def get_details_modal_form_class(
        open_incident_context: OpeningData,
        incident_type_value: str | None,
    ) -> type[SetIncidentDetails[Any]] | None:
        """Get the details modal form class based on the incident type.

        Returns None if no incident type is selected.
        """
        response_type = open_incident_context.get("response_type")
        if response_type is None:
            return None
        incident_types = INCIDENT_TYPES.get(response_type)
        if incident_types and len(incident_types) == 1:
            return incident_types[next(iter(incident_types.keys()))].get("slack_form")
        if incident_types and incident_type_value is not None:
            return incident_types[incident_type_value].get("slack_form")
        logger.debug(
            f"No incident type found for {open_incident_context}. No fallback."
        )
        return None

    def handle_modal_fn(  # type: ignore
        self,
        ack: Ack,
        body: dict[str, Any],
        user: User,
    ):
        """Handle response from /incident open modal."""
        data: OpeningData = json.loads(body["view"]["private_metadata"])

        details_form_data_raw = data.get("details_form_data", {})

        incident_type_value: str | None = data.get("incident_type", None)

        details_form_modal_class = self.get_details_modal_form_class(
            data, incident_type_value
        )
        if details_form_modal_class:
            details_form_class: type[CreateIncidentFormBase] = (
                details_form_modal_class.form_class
            )
            if details_form_class:
                details_form: CreateIncidentFormBase = details_form_class(
                    details_form_data_raw
                )
                details_form.is_valid()
                ack()
                try:
                    if hasattr(details_form, "trigger_incident_workflow") and callable(
                        details_form.trigger_incident_workflow
                    ):
                        details_form.trigger_incident_workflow(
                            creator=user,
                            impacts_data=data.get("impact_form_data") or {},
                        )
                except:  # noqa: E722
                    logger.exception("Error triggering incident workflow")
                    # XXX warn the user via DM!

    @app.action("incident_open_set_res_type_normal")
    @app.action("incident_open_set_res_type_critical")
    @staticmethod
    def handle_set_incident_response_type_action(
        ack: Ack, body: dict[str, Any]
    ) -> None:
        action_name: str = body.get("actions", [{}])[0].get("action_id", "")
        action_name = action_name.replace("incident_open_set_res_type_", "")
        opening_data = cast(
            "OpeningData", json.loads(body.get("actions", [{}])[0].get("value", {})) or {}
        )

        OpenModal._update_incident_modal(
            action_name, "response_type", ack, body, opening_data
        )

    @app.action("set_type")
    @staticmethod
    def handle_set_incident_type_action(ack: Ack, body: dict[str, Any]) -> None:
        incident_type_value: str = (
            body.get("actions", [{}])[0].get("selected_option", {}).get("value", "")
        )
        opening_data = json.loads(body.get("view", {}).get("private_metadata", "{}"))
        OpenModal._update_incident_modal(
            incident_type_value, "incident_type", ack, body, opening_data
        )

    @classmethod
    def _update_incident_modal(
        cls,
        action_value: str,
        metadata_key: Literal["incident_type", "response_type"],
        ack: Ack,
        body: dict[str, Any],
        opening_data: OpeningData,
    ) -> None:
        data: OpeningData = {**opening_data, metadata_key: action_value}  # type: ignore
        user = get_user_from_context(body)
        view = cls().build_modal_fn(open_incident_context=data, user=user)

        view.private_metadata = json.dumps(data, cls=SlackFormJSONEncoder)
        ack()
        update_modal(view=view, body=body)


modal_open = OpenModal()



from __future__ import annotations

import pytest

from firefighter.incidents.forms.select_impact import SelectImpactForm


@pytest.mark.django_db
@pytest.mark.parametrize(
    ("form_data", "expected_priority"),
    [
        (
            {
                "set_impact_type_business_impact": "HT",
                "set_impact_type_sellers_impact": "LT",
                "set_impact_type_customers_impact": "LT",
                "set_impact_type_employees_impact": "LT",
            },
            1,
        ),
        (
            {
                "set_impact_type_business_impact": "LO",
                "set_impact_type_sellers_impact": "LO",
                "set_impact_type_customers_impact": "HI",
                "set_impact_type_employees_impact": "LT",
            },
            2,
        ),
        (
            {
                "set_impact_type_business_impact": "LO",
                "set_impact_type_sellers_impact": "LO",
                "set_impact_type_customers_impact": "MD",
                "set_impact_type_employees_impact": "LT",
            },
            3,
        ),
        (
            {
                "set_impact_type_business_impact": "LO",
                "set_impact_type_sellers_impact": "LT",
                "set_impact_type_customers_impact": "LT",
                "set_impact_type_employees_impact": "LT",
            },
            4,
        ),
        (
            {
                "set_impact_type_business_impact": "LT",
                "set_impact_type_sellers_impact": "LT",
                "set_impact_type_customers_impact": "LT",
                "set_impact_type_employees_impact": "LT",
            },
            5,
        ),
    ],
)
def test_suggest_priority_from_impact(form_data, expected_priority):
    form = SelectImpactForm(data=form_data)
    assert form.is_valid()
    priority = form.suggest_priority_from_impact()
    assert priority == expected_priority


@pytest.mark.django_db
@pytest.mark.parametrize(
    ("form_data", "expected_priority"),
    [
        (
            {
                "set_impact_type_business_impact": "INVALID",
                "set_impact_type_sellers_impact": "LT",
                "set_impact_type_customers_impact": "LT",
                "set_impact_type_employees_impact": "LT",
            },
            6,
        ),
        (
            {
                "set_impact_type_business_impact": "LO",
                "set_impact_type_sellers_impact": "INVALID",
                "set_impact_type_customers_impact": "MD",
                "set_impact_type_employees_impact": "LT",
            },
            6,
        ),
        (
            {
                "set_impact_type_business_impact": "LO",
                "set_impact_type_sellers_impact": "LO",
                "set_impact_type_customers_impact": "INVALID",
                "set_impact_type_employees_impact": "LT",
            },
            6,
        ),
    ],
)
@pytest.mark.django_db
def test_suggest_priority_from_impact_with_invalid_form(form_data, expected_priority):
    form = SelectImpactForm(data=form_data)
    assert not form.is_valid()
    priority = form.suggest_priority_from_impact()
    assert priority == expected_priority
